from django.conf import settings
from django.core.management.base import BaseCommand

from ...storage_clients import client


class Command(BaseCommand):
    """Create a bucket based on the config."""

    help = "Create a bucket based on the config."

    def handle(self, *args, **options):
        bucket_name = settings.MINIO_STORAGE_MEDIA_BUCKET_NAME

        if client.client.bucket_exists(bucket_name):  # pragma: no cover
            self.stdout.write(f"Bucket: {bucket_name} already exists!")
        else:
            client.client.make_bucket(bucket_name)
            self.stdout.write(f"Created bucket: {bucket_name}")



from graphql_relay import (
    Connection,
    Edge,
    PageInfo,
    get_offset_with_default,
    offset_to_cursor,
)


def connection_from_array(
    data,
    args=None,
    connection_type=Connection,
    edge_type=Edge,
    page_info_type=PageInfo,
):
    """
    Replace graphql_relay.connection_from_array.

    This can be removed, when (or better if)
    https://github.com/graphql-python/graphql-relay-py/issues/12
    is resolved.

    A simple function that accepts an array and connection arguments, and returns
    a connection object for use in GraphQL. It uses array offsets as pagination,
    so pagination will only work if the array is static.
    """
    return connection_from_array_slice(
        data,
        args,
        slice_start=0,
        array_length=len(data),
        connection_type=connection_type,
        edge_type=edge_type,
        page_info_type=page_info_type,
    )


def connection_from_array_slice(
    array_slice,
    args=None,
    slice_start=0,
    array_length=None,
    array_slice_length=None,
    connection_type=Connection,
    edge_type=Edge,
    page_info_type=PageInfo,
):
    """
    Replace graphql_relay.connection_from_array_slice.

    This can be removed, when (or better if)
    https://github.com/graphql-python/graphql-relay-py/issues/12
    is resolved.

    Given a slice (subset) of an array, returns a connection object for use in
    GraphQL.
    This function is similar to `connectionFromArray`, but is intended for use
    cases where you know the cardinality of the connection, consider it too large
    to materialize the entire array, and instead wish pass in a slice of the
    total result large enough to cover the range specified in `args`.
    """
    args = args or {}
    before = args.get("before")
    after = args.get("after")
    first = args.get("first")
    last = args.get("last")
    if array_slice_length is None:  # pragma: no cover
        array_slice_length = len(array_slice)
    slice_end = slice_start + array_slice_length
    before_offset = get_offset_with_default(before, array_length)
    after_offset = get_offset_with_default(after, -1)

    start_offset = max(slice_start - 1, after_offset, -1) + 1
    end_offset = min(slice_end, before_offset, array_length)
    if isinstance(first, int):
        end_offset = min(end_offset, start_offset + first)
    if isinstance(last, int):
        start_offset = max(start_offset, end_offset - last)

    # If supplied slice is too large, trim it down before mapping over it.
    _slice = array_slice[
        max(start_offset - slice_start, 0) : array_slice_length
        - (slice_end - end_offset)
    ]
    edges = [
        edge_type(node=node, cursor=offset_to_cursor(start_offset + i))
        for i, node in enumerate(_slice)
    ]

    first_edge_cursor = edges[0].cursor if edges else None
    last_edge_cursor = edges[-1].cursor if edges else None

    return connection_type(
        edges=edges,
        page_info=page_info_type(
            start_cursor=first_edge_cursor,
            end_cursor=last_edge_cursor,
            has_previous_page=start_offset > 0,
            has_next_page=end_offset < array_length,
        ),
    )



import pytest

from caluma.caluma_workflow import models as workflow_models

from . import models


@pytest.fixture
def example_analytics(analytics_table, settings):
    analytics_table.starting_object = "cases"
    analytics_table.save()

    settings.META_FIELDS = ["foo"]

    analytics_table.fields.create(
        data_source="created_at",
        function=models.AnalyticsField.FUNCTION_VALUE,
        alias="created_at",
    )
    analytics_table.fields.create(
        data_source="created_at.quarter",
        function=models.AnalyticsField.FUNCTION_VALUE,
        alias="quarter",
    )
    analytics_table.fields.create(
        data_source="meta.foo",
        function=models.AnalyticsField.FUNCTION_VALUE,
        alias="foo",
    )
    analytics_table.fields.create(
        data_source="status",
        function=models.AnalyticsField.FUNCTION_VALUE,
        alias="statuuuuuus",
    )
    analytics_table.fields.create(
        data_source="document[top_form].top_question",
        function=models.AnalyticsField.FUNCTION_VALUE,
        alias="from_the_doc",
    )
    analytics_table.fields.create(
        data_source="document[top_form].form.sub_question",
        function=models.AnalyticsField.FUNCTION_VALUE,
        alias="sub_question_sumsumsum",
    )
    return analytics_table


@pytest.fixture
def analytics_cases(form_and_document, case_factory, set_date, work_item_factory):
    """Create a number of cases with documents to be used in testing.

    There are multiple cases with differing statuses and differing
    created_at values:
        * 3x RUNNING, created_at ranging from 2022-02-01 to 2022-01-03
        * 1x COMPLETED, created_at 2022-02-04
        * 1x SUSPENDED, created_at 2022-02-05
    """
    statuses = [
        # multiple cases with same status to test aggregates
        workflow_models.Case.STATUS_RUNNING,  # created_at 2022-02-01
        workflow_models.Case.STATUS_RUNNING,  # created_at 2022-02-02
        workflow_models.Case.STATUS_RUNNING,  # created_at 2022-02-03
        workflow_models.Case.STATUS_COMPLETED,  # created_at 2022-02-04
        workflow_models.Case.STATUS_SUSPENDED,  # created_at 2022-02-05
    ]

    # create each case with another modified_at date

    def _makecase(date, **kwargs):
        with set_date(date):
            _f, document, _q, _a = form_and_document(use_subform=True)
            case = case_factory(
                meta={"foo": "bar"},
                document=document,
                **kwargs,
            )
            work_item_factory(document=document.copy(), case=case)
        return case

    return [
        _makecase(f"2022-02-{day:02}", status=status)
        for day, status in enumerate(statuses, start=1)
    ]


@pytest.fixture
def example_pivot_table(example_analytics):

    status_field = example_analytics.fields.get(alias="statuuuuuus")

    status_field.function = models.AnalyticsField.FUNCTION_VALUE
    status_field.alias = "status"
    status_field.save()

    created_field = example_analytics.fields.get(alias="created_at")
    created_field.function = models.AnalyticsField.FUNCTION_MAX
    created_field.alias = "last_created"
    created_field.save()

    quarter_field = example_analytics.fields.get(alias="quarter")
    quarter_field.function = models.AnalyticsField.FUNCTION_MAX
    quarter_field.save()

    sub_q_field = example_analytics.fields.get(alias="sub_question_sumsumsum")
    sub_q_field.function = models.AnalyticsField.FUNCTION_SUM
    sub_q_field.alias = "sub_question_sumsumsum"
    sub_q_field.save()

    example_analytics.fields.all().exclude(
        pk__in=[status_field.pk, created_field.pk, quarter_field.pk, sub_q_field.pk]
    ).delete()

    return example_analytics



from django.contrib.postgres.fields import ArrayField
from django.db import models
from django.db.models.constraints import UniqueConstraint
from localized_fields.fields import LocalizedField

from caluma.caluma_core.models import SlugModel, UUIDModel

from . import pivot_table, simple_table


class AnalyticsTable(SlugModel):
    STARTING_OBJECT_CHOICES = simple_table.BaseStartingObject.as_choices()

    meta = models.JSONField(default=dict)

    disable_visibilities = models.BooleanField(default=False)
    name = LocalizedField(blank=False, null=False, required=False)

    starting_object = models.CharField(max_length=250, choices=STARTING_OBJECT_CHOICES)

    def is_extraction(self):
        # We are an extraction table if there are ONLY
        # `FUNCTION_VALUE` type fields.
        return not self.fields.exclude(function=AnalyticsField.FUNCTION_VALUE).exists()

    def get_starting_object(self, info):
        return simple_table.BaseStartingObject.get_object(self.starting_object, info)

    def get_analytics(self, info):
        return (
            simple_table.SimpleTable(self, info)
            if self.is_extraction()
            else pivot_table.PivotTable(self, info)
        )

    def __repr__(self):
        return f"AnalyticsTable<{self.slug}>"


class AnalyticsField(UUIDModel):
    """Analytics field.

    May represent an aggregate (pivot) field on a pivot table, or
    an extraction field on an extraction table.

    The data_source references a field listed from the available fields
    that the starting object of the table provides.

    The function defines how the data of the
    specified source field is aggregated. While most of the functions
    are obvious by their name, the `FUNCTION_VALUE` is special: Those
    fields will not be aggregated, but the other aggregates will be
    grouped by this value.

    For example, if you define the following fields:
        - foo SUM
        - bar MIN
        - baz VALUE
    The output will contain one row for each unique value of `baz`.
    The sums and minimums calculated will all be in relation to rows with
    the same `baz` value.
    """

    FUNCTION_VALUE = "value"
    FUNCTION_SUM = "sum"
    FUNCTION_AVERAGE = "avg"
    FUNCTION_MAX = "max"
    FUNCTION_COUNT = "count"
    FUNCTION_MIN = "min"

    FUNCTION_CHOICES = [
        (FUNCTION_VALUE, "Plain value (Implies GROUP BY)"),
        (FUNCTION_SUM, "Sum"),
        (FUNCTION_COUNT, "Count"),
        (FUNCTION_AVERAGE, "Average (Mean)"),
        (FUNCTION_MAX, "Max"),
        (FUNCTION_MIN, "Min"),
    ]

    alias = models.CharField(max_length=100)
    meta = models.JSONField(default=dict)

    data_source = models.TextField()
    table = models.ForeignKey(
        AnalyticsTable, on_delete=models.CASCADE, related_name="fields"
    )
    filters = ArrayField(models.TextField(), blank=True, null=True, default=list)

    function = models.CharField(
        max_length=20, choices=FUNCTION_CHOICES, default=FUNCTION_VALUE
    )

    show_output = models.BooleanField(default=True)
    sort = models.PositiveIntegerField(editable=False, db_index=True, default=0)

    def __repr__(self):
        return f"AnalyticsField<{self.table.slug}.{self.alias}>"

    class Meta:
        ordering = ["sort", "-created_at"]
        constraints = [
            UniqueConstraint(
                name="unique_data_source",
                fields=["table", "data_source", "function"],
            ),
            UniqueConstraint(
                name="unique_alias",
                fields=["table", "alias"],
            ),
        ]



# Generated by Django 2.2.6 on 2019-10-25 09:50

import uuid

import django.db.models.deletion
import localized_fields.fields.field
import simple_history.models
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("caluma_form", "0024_auto_20190919_1244")]

    operations = [
        migrations.CreateModel(
            name="HistoricalDynamicOption",
            fields=[
                ("created_at", models.DateTimeField(blank=True, editable=False)),
                ("modified_at", models.DateTimeField(blank=True, editable=False)),
                (
                    "created_by_user",
                    models.CharField(blank=True, max_length=150, null=True),
                ),
                (
                    "created_by_group",
                    models.CharField(
                        blank=True, db_index=True, max_length=150, null=True
                    ),
                ),
                ("history_user_id", models.CharField(max_length=150, null=True)),
                (
                    "id",
                    models.UUIDField(db_index=True, default=uuid.uuid4, editable=False),
                ),
                ("value", models.CharField(blank=True, max_length=255, null=True)),
                ("label", localized_fields.fields.field.LocalizedField(required=[])),
                (
                    "history_id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("history_date", models.DateTimeField()),
                ("history_change_reason", models.CharField(max_length=100, null=True)),
                (
                    "history_type",
                    models.CharField(
                        choices=[("+", "Created"), ("~", "Changed"), ("-", "Deleted")],
                        max_length=1,
                    ),
                ),
                (
                    "document",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="caluma_form.Document",
                    ),
                ),
                (
                    "question",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="caluma_form.Question",
                    ),
                ),
            ],
            options={
                "verbose_name": "historical dynamic option",
                "ordering": ("-history_date", "-history_id"),
                "get_latest_by": "history_date",
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name="DynamicOption",
            fields=[
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("modified_at", models.DateTimeField(auto_now=True)),
                (
                    "created_by_user",
                    models.CharField(blank=True, max_length=150, null=True),
                ),
                (
                    "created_by_group",
                    models.CharField(
                        blank=True, db_index=True, max_length=150, null=True
                    ),
                ),
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("value", models.CharField(blank=True, max_length=255, null=True)),
                ("label", localized_fields.fields.field.LocalizedField(required=[])),
                (
                    "document",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="caluma_form.Document",
                    ),
                ),
                (
                    "question",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="caluma_form.Question",
                    ),
                ),
            ],
            options={"abstract": False},
        ),
    ]



# Generated by Django 3.2.15 on 2022-10-24 11:10

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("caluma_analytics", "0004_simple_history"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="analyticsfield",
            options={"ordering": ["sort", "-created_at"]},
        ),
        migrations.AddField(
            model_name="analyticsfield",
            name="sort",
            field=models.PositiveIntegerField(db_index=True, default=0, editable=False),
        ),
        migrations.AddField(
            model_name="historicalanalyticsfield",
            name="sort",
            field=models.PositiveIntegerField(db_index=True, default=0, editable=False),
        ),
    ]



from typing import Any, Tuple

from django import forms
from django.db.models import OuterRef, Subquery
from django.db.models.expressions import F
from django.db.models.query import QuerySet
from rest_framework import exceptions

from caluma.caluma_core.ordering import CalumaOrdering, OrderingFieldType
from caluma.caluma_form.models import Answer, Question


class AnswerValueOrdering(CalumaOrdering):
    """Ordering filter for ordering documents by value of an answer.

    This can also be used for ordering other things that contain documents,
    such as cases, work items etc.
    """

    field_class = forms.CharField

    def __init__(self, document_via=None, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._document_locator_prefix = (
            "" if document_via is None else f"{document_via}__"
        )

    def get_ordering_value(
        self, qs: QuerySet, value: Any
    ) -> Tuple[QuerySet, OrderingFieldType]:
        # First, join the requested answer, then annotate the QS accordingly.
        # Last, return a field corresponding to the value
        question = Question.objects.get(pk=value)
        QUESTION_TYPE_TO_FIELD = {
            Question.TYPE_INTEGER: "value",
            Question.TYPE_FLOAT: "value",
            Question.TYPE_DATE: "date",
            Question.TYPE_CHOICE: "value",
            Question.TYPE_TEXTAREA: "value",
            Question.TYPE_TEXT: "value",
            Question.TYPE_FILES: "files__name",
            Question.TYPE_DYNAMIC_CHOICE: "value",
        }

        try:
            value_field = QUESTION_TYPE_TO_FIELD[question.type]
        except KeyError:  # pragma: no cover
            raise exceptions.ValidationError(
                f"Question '{question.slug}' has unsupported type {question.type} for ordering"
            )

        answers_subquery = Subquery(
            Answer.objects.filter(
                question=question,
                document=OuterRef(f"{self._document_locator_prefix}pk"),
            ).values(value_field)
        )
        ann_name = f"order_{value}"

        qs = qs.annotate(**{ann_name: answers_subquery})

        # TODO: respect document_via
        return qs, F(ann_name)



from django_filters.rest_framework import FilterSet, MultipleChoiceFilter

from ..caluma_core.filters import MetaFilterSet, SearchFilter
from ..caluma_core.ordering import AttributeOrderingFactory, MetaFieldOrdering
from . import models


class AnalyticsTableFilterSet(MetaFilterSet):
    search = SearchFilter(fields=("slug", "name"))
    slugs = MultipleChoiceFilter(field_name="slug")

    class Meta:
        model = models.AnalyticsTable
        fields = (
            "slug",
            "name",
        )


class AnalyticsFieldFilterSet(MetaFilterSet):
    search = SearchFilter(
        fields=(
            "alias",
            "data_source",
            "function",
            "table__name",
            "source_field__alias",
        )
    )
    slugs = MultipleChoiceFilter(field_name="slug")

    class Meta:
        model = models.AnalyticsField
        fields = (
            "alias",
            "table",
        )


class AnalyticsFieldOrderSet(FilterSet):
    meta = MetaFieldOrdering()
    attribute = AttributeOrderingFactory(
        models.AnalyticsField,
        fields=["created_at", "modified_at", "alias"],
    )

    class Meta:
        model = models.AnalyticsField
        fields = ("meta", "attribute")


class AnalyticsTableOrderSet(FilterSet):
    meta = MetaFieldOrdering()
    attribute = AttributeOrderingFactory(
        models.AnalyticsTable,
        fields=["created_at", "modified_at", "slug", "name"],
    )

    class Meta:
        model = models.AnalyticsTable
        fields = ("meta", "attribute")



# Generated by Django 2.2.14 on 2020-07-30 11:35

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("caluma_workflow", "0022_workitem_name_description")]

    operations = [
        migrations.AddField(
            model_name="historicalworkitem",
            name="previous_work_item",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="caluma_workflow.WorkItem",
            ),
        ),
        migrations.AddField(
            model_name="workitem",
            name="previous_work_item",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="succeeding_work_items",
                to="caluma_workflow.WorkItem",
            ),
        ),
    ]



import enum
from functools import reduce, singledispatch

import graphene
from django import forms
from django.contrib.postgres.fields.hstore import KeyTransform
from django.contrib.postgres.search import SearchVector
from django.db import models
from django.db.models.constants import LOOKUP_SEP
from django.db.models.expressions import OrderBy
from django.db.models.fields.json import KeyTextTransform
from django.db.models.functions import Cast
from django.utils import translation
from django_filters.conf import settings as filters_settings
from django_filters.constants import EMPTY_VALUES
from django_filters.fields import ChoiceField
from django_filters.rest_framework import (
    CharFilter,
    DateTimeFilter,
    Filter,
    FilterSet,
    MultipleChoiceFilter,
)
from graphene import Enum, InputObjectType, List
from graphene.types import generic
from graphene.types.utils import get_type
from graphene.utils.str_converters import to_camel_case
from graphene_django import filter
from graphene_django.converter import convert_choice_name
from graphene_django.filter.filterset import GrapheneFilterSetMixin
from graphene_django.forms.converter import convert_form_field
from graphene_django.registry import get_global_registry
from localized_fields.fields import LocalizedField

from .forms import GlobalIDFormField, GlobalIDMultipleChoiceField
from .ordering import CalumaOrdering
from .relay import extract_global_id
from .types import DjangoConnectionField


class CompositeFieldClass(forms.MultiValueField):
    """Mixin to build complex field classes.

    This is just to pretend to Graphene that it's a composite type.
    It's the base of the internal representation that only passes
    values from the request down to the filters (or similar).

    The actual schema type is generated via the `convert_form_field()`
    function from `graphene_django.forms.converter`.
    """

    def __init__(self, label, *, fields=None, **kwargs):
        fields = (forms.CharField(), forms.CharField())
        super().__init__(fields=fields)

    def clean(self, data):
        # override parent clean() which would reject our data structure.
        # We don't validate, as the structure is already enforced by the
        # schema.

        return data


class AscDesc(Enum):
    ASC = "ASC"
    DESC = "DESC"


class FilterCollectionFilter(Filter):
    def filter(self, qs, value):
        if value in EMPTY_VALUES:
            return qs

        filter_coll = self.filterset_class()
        for flt in value:
            if not flt:
                continue
            invert = flt.pop("invert", False)
            flt_key = list(flt.keys())[0]
            flt_val = flt[flt_key]
            filter = filter_coll.filters[flt_key]

            new_qs = filter.filter(qs, flt_val)

            if invert:
                qs = qs.exclude(pk__in=new_qs)
            else:
                qs = new_qs

        return qs


class FilterCollectionOrdering(Filter):
    def _order_part(self, qs, ord, filter_coll):
        direction = ord.pop("direction", "ASC")

        assert len(ord) == 1
        filt_name = list(ord.keys())[0]
        filter = filter_coll.filters[filt_name]
        qs, field = filter.get_ordering_value(qs, ord[filt_name])

        # Normally, people use ascending order, and in this context it seems
        # natural to have NULL entries at the end.
        # Making the `nulls_first`/`nulls_last` parameter accessible in the
        # GraphQL interface would be overkill, at least for now.
        return (
            qs,
            OrderBy(
                field,
                descending=(direction == "DESC"),
                nulls_first=(direction == "DESC"),
                nulls_last=(direction == "ASC"),
            ),
        )

    def filter(self, qs, value):
        if value in EMPTY_VALUES:
            return qs
        filter_coll = self.filterset_class()

        order_by = []
        for ord in value:
            qs, order_field = self._order_part(qs, ord, filter_coll)
            order_by.append(order_field)

        if order_by:
            qs = qs.order_by(*order_by)

        return qs


def FilterCollectionFactory(filterset_class, ordering):  # noqa:C901
    """
    Build a single filter from a `FilterSet`.

    This converts an arbitrary `FilterSet` class into a single filter that
    allows chaining of filters as a list. In addition, this introduces
    an optional `invert` parameter, allowing requestors to use
    a filter for either inclusion or exclusion.

    On the schema side, this generates a new type to represent the filter value
    whose name is derived from the given filterset class.

    Usage:
    >>> class MyFilterSet(FilterSet):
    ...     filter = FilterCollectionFactory(FilterSetWithActualFilters, ordering=False)

    """

    field_class_name = f"{filterset_class.__name__}Field"
    field_type_name = f"{filterset_class.__name__}Type"
    collection_name = f"{filterset_class.__name__}Collection"

    # The field class.
    custom_field_class = type(field_class_name, (CompositeFieldClass,), {})

    @convert_form_field.register(custom_field_class)
    def convert_field(field):

        registry = get_global_registry()
        converted = registry.get_converted_field(field)
        if converted:
            return converted

        _filter_coll = filterset_class()

        def _get_or_make_field(name, filt):
            return convert_form_field(filt.field)

        def _should_include_filter(filt):
            # if we're in ordering mode, we want
            # to return True for all CalumaOrdering types,
            # and if it's false, we want the opposite
            return ordering == isinstance(filt, CalumaOrdering)

        filter_fields = {
            name: _get_or_make_field(name, filt)
            for name, filt in _filter_coll.filters.items()
            if _should_include_filter(filt)
        }

        if ordering:
            filter_fields["direction"] = AscDesc(default=AscDesc.ASC, required=False)
        else:
            filter_fields["invert"] = graphene.Boolean(required=False, default=False)

        filter_type = type(field_type_name, (InputObjectType,), filter_fields)

        converted = List(filter_type)
        registry.register_converted_field(field, converted)
        return converted

    filter_impl = FilterCollectionOrdering if ordering else FilterCollectionFilter

    filter_coll = type(
        collection_name,
        (filter_impl,),
        {"field_class": custom_field_class, "filterset_class": filterset_class},
    )
    return filter_coll()


def CollectionFilterSetFactory(filterset_class, orderset_class=None):
    """
    Build single-filter filterset classes.

    Use this in place of a regular filterset_class parametrisation in
    the serializers.
    If you pass the optional `orderset_class` parameter, it is used for
    an `order` filter. The filters defined in the `orderset_class` must
    inherit from `caluma.caluma_core.ordering.CalumaOrdering` and provide a
    `get_ordering_value()` method.

    Example:
    >>> all_documents = DjangoFilterConnectionField(
    ...    Document, filterset_class=CollectionFilterSetFactory(DocumentFilterSet)
    ... )

    """

    suffix = "" if orderset_class else "NoOrdering"
    cache_key = filterset_class.__name__ + suffix

    if cache_key in CollectionFilterSetFactory._cache:
        return CollectionFilterSetFactory._cache[cache_key]

    coll_fields = {"filter": FilterCollectionFactory(filterset_class, ordering=False)}
    if orderset_class:
        coll_fields["order"] = FilterCollectionFactory(orderset_class, ordering=True)

    ret = CollectionFilterSetFactory._cache[cache_key] = type(
        f"{filterset_class.__name__}Collection",
        (FilterSet,),
        {
            **coll_fields,
            "Meta": type(
                "Meta",
                (filterset_class.Meta,),
                {
                    "model": filterset_class.Meta.model,
                    "fields": tuple(coll_fields.keys()),
                },
            ),
        },
    )

    return ret


CollectionFilterSetFactory._cache = {}


class GlobalIDFilter(Filter):
    field_class = GlobalIDFormField

    def filter(self, qs, value):
        _id = None
        if value is not None:
            _id = extract_global_id(value)
        return super(GlobalIDFilter, self).filter(qs, _id)


class GlobalIDMultipleChoiceFilter(MultipleChoiceFilter):
    field_class = GlobalIDMultipleChoiceField

    def filter(self, qs, value):
        gids = [extract_global_id(v) for v in value]
        return super(GlobalIDMultipleChoiceFilter, self).filter(qs, gids)


class LocalizedFilter(Filter):
    def filter(self, qs, value):
        if value in EMPTY_VALUES:  # pragma: no cover
            return qs

        lang = translation.get_language()
        filter_expr = "{0}__{1}__{2}".format(self.field_name, lang, self.lookup_expr)
        return qs.filter(**{filter_expr: value})


GrapheneFilterSetMixin.FILTER_DEFAULTS.update(
    {
        LocalizedField: {"filter_class": LocalizedFilter},
        models.AutoField: {"filter_class": GlobalIDFilter},
        models.OneToOneField: {"filter_class": GlobalIDFilter},
        models.ForeignKey: {"filter_class": GlobalIDFilter},
        models.ManyToManyField: {"filter_class": GlobalIDMultipleChoiceFilter},
        models.ManyToOneRel: {"filter_class": GlobalIDMultipleChoiceFilter},
        models.ManyToManyRel: {"filter_class": GlobalIDMultipleChoiceFilter},
    }
)


class SearchFilter(Filter):
    """
    Enable fulltext search on queryset.

    Define fields which need to be searched in.
    """

    def __init__(self, *args, fields, **kwargs):
        self.fields = fields
        super().__init__(*args, **kwargs)

    def _get_model_field(self, model, field):
        model_field = model._meta.get_field(field)
        return model_field, getattr(model_field, "related_model", None)

    def _build_search_expression(self, field_lookup):
        # TODO: is there no Django API which allows conversion of lookup to django field?
        model_field, _ = reduce(
            lambda model_tuple, field: self._get_model_field(model_tuple[1], field),
            field_lookup.split(LOOKUP_SEP),
            (None, self.model),
        )

        if isinstance(model_field, LocalizedField):
            lang = translation.get_language()
            return KeyTransform(lang, field_lookup)
        elif isinstance(model_field, models.JSONField):
            return Cast(field_lookup, models.TextField())

        return field_lookup

    def filter(self, qs, value):
        if value in EMPTY_VALUES:  # pragma: no cover
            return qs

        qs = qs.annotate(
            search=SearchVector(
                *[self._build_search_expression(field) for field in self.fields]
            )
        )

        return qs.filter(search=value)


class ListField(forms.Field):
    """List field as to allow actual lists in ordering vs csv string."""

    pass


class IntegerFilter(Filter):
    field_class = forms.IntegerField


class BaseFilterSet(GrapheneFilterSetMixin, FilterSet):
    created_by_user = CharFilter()
    created_by_group = CharFilter()
    modified_by_user = CharFilter()
    modified_by_group = CharFilter()

    created_before = DateTimeFilter(
        field_name="created_at",
        lookup_expr="lt",
        label="Only return entries created before the given DateTime (exclusive)",
    )
    created_after = DateTimeFilter(
        field_name="created_at",
        lookup_expr="gte",
        label="Only return entries created at or after the given DateTime (inclusive)",
    )


class JSONLookupMode(Enum):
    EXACT = "exact"
    STARTSWITH = "startswith"
    CONTAINS = "contains"
    ICONTAINS = "icontains"
    IN = "in"
    GTE = "gte"
    GT = "gt"
    LTE = "lte"
    LT = "lt"


class JSONValueFilterType(InputObjectType):
    key = graphene.String(required=True)
    value = generic.GenericScalar(required=True)
    lookup = JSONLookupMode()


class JSONValueFilterField(CompositeFieldClass):
    pass


class JSONValueFilter(Filter):
    field_class = JSONValueFilterField

    def __init__(self, *args, lookup_expr=None, **kwargs):
        if lookup_expr is None:
            lookup_expr = JSONLookupMode.get(filters_settings.DEFAULT_LOOKUP_EXPR)
        super().__init__(*args, lookup_expr=lookup_expr, **kwargs)

    def filter(self, qs, value):
        if value in EMPTY_VALUES:  # pragma: no cover
            return qs

        for expr in value:
            if expr in EMPTY_VALUES:  # pragma: no cover
                continue

            lookup = expr.get("lookup") or self.lookup_expr
            lookup_expr = (hasattr(lookup, "value") and lookup.value) or lookup

            # "contains" behaves differently on JSONFields as it does on TextFields.
            # That's why we annotate the queryset with the value.
            # Some discussion about it can be found here:
            # https://code.djangoproject.com/ticket/26511
            if isinstance(expr["value"], str):
                qs = qs.annotate(
                    field_val=Cast(
                        KeyTextTransform(expr["key"], self.field_name),
                        models.CharField(),
                    ),
                )
                lookup = {f"field_val__{lookup_expr}": expr["value"]}
            else:
                lookup = {
                    f"{self.field_name}__{expr['key']}__{lookup_expr}": expr["value"]
                }
            qs = qs.filter(**lookup)
        return qs

    @staticmethod
    @convert_form_field.register(JSONValueFilterField)
    def convert_meta_value_field(field):
        converted = List(JSONValueFilterType)
        get_global_registry().register_converted_field(field, converted)
        return converted


class MetaFilterSet(BaseFilterSet):
    meta_has_key = CharFilter(lookup_expr="has_key", field_name="meta")
    meta_value = JSONValueFilter(field_name="meta")


class DjangoFilterConnectionField(
    filter.DjangoFilterConnectionField, DjangoConnectionField
):
    @property
    def filterset_class(self):
        return self._provided_filterset_class

    @classmethod
    def connection_resolver(
        cls,
        resolver,
        connection,
        default_manager,
        queryset_resolver,
        max_limit,
        enforce_first_or_last,
        root,
        info,
        **args,
    ):
        return super().connection_resolver(
            resolver=resolver,
            connection=connection,
            default_manager=default_manager,
            queryset_resolver=queryset_resolver,
            max_limit=max_limit,
            enforce_first_or_last=enforce_first_or_last,
            root=root,
            info=info,
            **cls._clean_args_for_queryset_resolver(args),
        )

    @classmethod
    def _clean_args_for_queryset_resolver(cls, args):
        # Graphene parses incoming data into Enums too early, thus our filters
        # will receive enum objects that cannot be parsed
        #
        # TODO: check if this is still required after the below
        # resolve_queryset() is completely implemented (we assumed it's
        # the Enums, but it was actually the list in order_by. We'll keep
        # it here until we KNOW we can remove it again..)
        @singledispatch
        def clean(data):
            return data

        @clean.register(enum.Enum)
        def _(data):
            return data.value

        @clean.register(list)
        def _(data):
            return [clean(e) for e in data]

        @clean.register(dict)
        def _(data):
            return {k: clean(v) for k, v in data.items()}

        return clean(args)


class DjangoFilterInterfaceConnectionField(DjangoFilterConnectionField):
    """Filter connection field for abstract interface types like Answer, Question and Task."""

    @property
    def model(self):
        return self.filterset_class._meta.model

    @property
    def type(self):
        return get_type(self._type)


@convert_form_field.register(forms.ChoiceField)
@convert_form_field.register(ChoiceField)
def convert_choice_field_to_enum(field):
    """
    Add support to convert ordering choices to Graphql enum.

    Label is used as enum name.
    """

    def get_choices(choices):
        for value, help_text in choices:
            if value:
                name = convert_choice_name(value)
                description = help_text
                yield name, value, description

    name = to_camel_case(field.label)
    choices = list(get_choices(field.choices))
    named_choices = [(c[0], c[1]) for c in choices]
    named_choices_descriptions = {c[0]: c[2] for c in choices}

    class EnumWithDescriptionsType(object):
        @property
        def description(self):
            return named_choices_descriptions[self.name]

    enum = Enum(name, list(named_choices), type=EnumWithDescriptionsType)
    converted = enum(description=field.help_text, required=field.required)

    get_global_registry().register_converted_field(field.label, converted)
    return converted


def generate_list_filter_class(inner_type):
    """
    Return a Filter class that will resolve into a List(`inner_type`) graphene type.

    This allows us to do things like use `__in` and `__overlap` filters that accept
    graphene lists instead of a comma delimited value string that's interpolated into
    a list by django_filters.BaseCSVFilter (which is used to define
    django_filters.BaseInFilter)
    """

    form_field = type(f"List{inner_type.__name__}FormField", (forms.Field,), {})
    filter_class = type(
        f"{inner_type.__name__}ListFilter",
        (Filter,),
        {
            "field_class": form_field,
            "__doc__": (
                f"{inner_type.__name__}ListFilter is a small extension of a raw "
                f"django_filters.Filter that allows us to express graphql "
                f"List({inner_type.__name__}) arguments using FilterSets. "
                f"Note that the given values are passed directly into queryset filters."
            ),
        },
    )

    def do_convert_type(x):
        return graphene.List(inner_type, required=x.required)

    convert_form_field.register(form_field)(do_convert_type)

    return filter_class


StringListFilter = generate_list_filter_class(graphene.String)



from django.core.management.base import BaseCommand, CommandError

from caluma.caluma_form.api import copy_form
from caluma.caluma_form.models import Form


class Command(BaseCommand):
    """Copy a form."""

    help = "Copy a form."

    def add_arguments(self, parser):
        parser.add_argument(
            "source",
            type=str,
            help="The slug of the form to copy.",
        )
        parser.add_argument(
            "--slug",
            "-s",
            dest="slug",
            type=str,
            required=True,
            help="The slug of the newly created form.",
        )
        parser.add_argument(
            "--name",
            "-n",
            dest="name",
            type=str,
            default="",
            help="The name of the newly created form.",
        )
        parser.add_argument(
            "--description",
            "-d",
            dest="description",
            type=str,
            default="",
            help="The description of the newly created form.",
        )
        parser.add_argument(
            "--is-published",
            "-p",
            dest="is_published",
            type=bool,
            default=False,
            help="Whether the newly created form is published.",
        )

    def handle(self, *args, **options):
        source_slug = options["source"]
        new_slug = options["slug"]

        try:
            source = Form.objects.get(pk=source_slug)
        except Form.DoesNotExist:
            raise CommandError(f"Form '{source_slug}' not found - can't copy")

        copy_form(
            source=source,
            slug=new_slug,
            name=options["name"],
            description=options["description"],
            is_published=options["is_published"],
        )

        self.stdout.write(f"Copied form '{source_slug}' to new form '{new_slug}'")



# Generated by Django 3.2.13 on 2022-06-09 11:26

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("caluma_form", "0044_migrate_format_validators"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="historicalanswer",
            options={
                "get_latest_by": ("history_date", "history_id"),
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical answer",
                "verbose_name_plural": "historical answers",
            },
        ),
        migrations.AlterModelOptions(
            name="historicalanswerdocument",
            options={
                "get_latest_by": ("history_date", "history_id"),
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical answer document",
                "verbose_name_plural": "historical answer documents",
            },
        ),
        migrations.AlterModelOptions(
            name="historicaldocument",
            options={
                "get_latest_by": ("history_date", "history_id"),
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical document",
                "verbose_name_plural": "historical documents",
            },
        ),
        migrations.AlterModelOptions(
            name="historicaldynamicoption",
            options={
                "get_latest_by": ("history_date", "history_id"),
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical dynamic option",
                "verbose_name_plural": "historical dynamic options",
            },
        ),
        migrations.AlterModelOptions(
            name="historicalfile",
            options={
                "get_latest_by": ("history_date", "history_id"),
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical file",
                "verbose_name_plural": "historical files",
            },
        ),
        migrations.AlterModelOptions(
            name="historicalform",
            options={
                "get_latest_by": ("history_date", "history_id"),
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical form",
                "verbose_name_plural": "historical forms",
            },
        ),
        migrations.AlterModelOptions(
            name="historicalformquestion",
            options={
                "get_latest_by": ("history_date", "history_id"),
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical form question",
                "verbose_name_plural": "historical form questions",
            },
        ),
        migrations.AlterModelOptions(
            name="historicaloption",
            options={
                "get_latest_by": ("history_date", "history_id"),
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical option",
                "verbose_name_plural": "historical options",
            },
        ),
        migrations.AlterModelOptions(
            name="historicalquestion",
            options={
                "get_latest_by": ("history_date", "history_id"),
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical question",
                "verbose_name_plural": "historical questions",
            },
        ),
        migrations.AlterModelOptions(
            name="historicalquestionoption",
            options={
                "get_latest_by": ("history_date", "history_id"),
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical question option",
                "verbose_name_plural": "historical question options",
            },
        ),
        migrations.AlterField(
            model_name="historicalanswer",
            name="history_date",
            field=models.DateTimeField(db_index=True),
        ),
        migrations.AlterField(
            model_name="historicalanswerdocument",
            name="history_date",
            field=models.DateTimeField(db_index=True),
        ),
        migrations.AlterField(
            model_name="historicaldocument",
            name="history_date",
            field=models.DateTimeField(db_index=True),
        ),
        migrations.AlterField(
            model_name="historicaldynamicoption",
            name="history_date",
            field=models.DateTimeField(db_index=True),
        ),
        migrations.AlterField(
            model_name="historicalfile",
            name="history_date",
            field=models.DateTimeField(db_index=True),
        ),
        migrations.AlterField(
            model_name="historicalform",
            name="history_date",
            field=models.DateTimeField(db_index=True),
        ),
        migrations.AlterField(
            model_name="historicalformquestion",
            name="history_date",
            field=models.DateTimeField(db_index=True),
        ),
        migrations.AlterField(
            model_name="historicaloption",
            name="history_date",
            field=models.DateTimeField(db_index=True),
        ),
        migrations.AlterField(
            model_name="historicalquestion",
            name="history_date",
            field=models.DateTimeField(db_index=True),
        ),
        migrations.AlterField(
            model_name="historicalquestionoption",
            name="history_date",
            field=models.DateTimeField(db_index=True),
        ),
    ]



import inspect

from django.core.exceptions import ImproperlyConfigured

from .collections import list_duplicates


def validation_for(mutation):
    """Decorate function to overwriting validation of specific mutation."""

    def decorate(fn):
        if not hasattr(fn, "_validations"):
            fn._validations = []

        fn._validations.append(mutation)
        return fn

    return decorate


class BaseValidation(object):
    """Basic validation class to be extended by any validation implementation.

    In combination with the decorators `@validation_for` a custom validation class
    can define validations for specific mutations.
    It is also possible to execute the same validation logic for all mutations
    using the base class `Mutation`.

    A custom validation class could look like this:
    ```
    >>> from caluma.caluma_form.schema import SaveForm
    ... from caluma.caluma_core.mutation import Mutation
    ... from rest_framework import exceptions
    ...
    ... class CustomValidation(BaseValidation):
    ...     @validation_for(Mutation)
    ...     def validate_mutation(self, mutation, data, info):
    ...         # add your default specific validation code here
    ...         return data
    ...
    ...     @validation_for(SaveForm)
    ...     def validate_save_form(self, mutation, data, info):
    ...         if data['meta'] and 'admin' not in info.context.user.groups:
    ...             raise exceptions.ValidationException('May not change meta on form')
    ...         return data
    """

    def __init__(self):
        validation_fns = inspect.getmembers(self, lambda m: hasattr(m, "_validations"))

        validation_muts = [
            mutation.__name__
            for _, fn in validation_fns
            for mutation in fn._validations
        ]
        validation_muts_dups = list_duplicates(validation_muts)
        if validation_muts_dups:
            raise ImproperlyConfigured(
                f"`validation_for` defined multiple times for "
                f"{', '.join(validation_muts_dups)} in {str(self)}"
            )
        self._validations = {
            mutation: fn for _, fn in validation_fns for mutation in fn._validations
        }

    def validate(self, mutation, data, info):
        for cls in mutation.mro():
            if cls in self._validations:
                return self._validations[cls](mutation, data, info)
        return data



# Generated by Django 2.2.10 on 2020-02-20 13:11

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("caluma_form", "0031_auto_20200220_0910")]

    operations = [
        migrations.AddField(
            model_name="document",
            name="source",
            field=models.ForeignKey(
                blank=True,
                help_text="Reference this document has been copied from",
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="copies",
                to="caluma_form.Document",
            ),
        ),
        migrations.AddField(
            model_name="historicaldocument",
            name="source",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                help_text="Reference this document has been copied from",
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="caluma_form.Document",
            ),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.18 on 2019-02-08 13:02
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("caluma_workflow", "0007_auto_20190206_0854")]

    operations = [
        migrations.RemoveField(model_name="workflow", name="start"),
        migrations.AddField(
            model_name="workflow",
            name="start_tasks",
            field=models.ManyToManyField(
                help_text="Starting task(s) of the workflow.",
                related_name="_workflow_start_tasks_+",
                to="caluma_workflow.Task",
            ),
        ),
    ]



import graphene
from django.shortcuts import get_object_or_404
from graphene import relay
from graphene.types import ObjectType, generic
from graphene_django.rest_framework import serializer_converter

from ..caluma_core.filters import (
    CollectionFilterSetFactory,
    DjangoFilterConnectionField,
    DjangoFilterInterfaceConnectionField,
)
from ..caluma_core.mutation import Mutation, UserDefinedPrimaryKeyMixin
from ..caluma_core.relay import extract_global_id
from ..caluma_core.types import (
    ConnectionField,
    CountableConnectionBase,
    DjangoObjectType,
    Node,
    enum_type_from_field,
)
from ..caluma_data_source.data_source_handlers import get_data_source_data
from ..caluma_data_source.schema import DataSourceDataConnection
from . import filters, models, serializers
from .format_validators import get_format_validators
from .validators import get_document_validity


def resolve_answer(answer):
    if answer.question.type in [
        models.Question.TYPE_STATIC,
        models.Question.TYPE_ACTION_BUTTON,
    ]:
        raise Exception(
            'Questions of type "static" and "action_button" should never have an answer!'
        )
    return QUESTION_ANSWER_TYPES[answer.question.type]


def resolve_question(question):
    return QUESTION_OBJECT_TYPES[question.type]


class FormDjangoObjectType(DjangoObjectType):
    """
    Django object type with improved `is_type_of`.

    This is needed to resolve our different Answer and Question types.
    """

    @classmethod
    def is_type_of(cls, root, info):
        is_type = super().is_type_of(root, info)
        if is_type:
            if root._meta.model == models.Answer:
                return resolve_answer(root)._meta.name == cls._meta.name
            if root._meta.model == models.Question:
                return resolve_question(root)._meta.name == cls._meta.name
        return is_type

    class Meta:
        abstract = True


class QuestionJexl(graphene.String):
    """Question jexl expression returning boolean.

    Following transform can be used:
    * answer - access answer of document by question slug
    * mapby - map list by key. Helpful to work with table answers
      whereas an answer is a list of dicts.

    Following context is available:
    * form - access form of document

    Examples:
    * 'answer' == 'question-slug'|answer
    * 'answer' in 'list-question-slug'|answer
    * 'answer' in 'table-question-slug'|answer|mapby('column-question')
    * 'form-slug' == form

    """

    pass


serializer_converter.get_graphene_type_from_serializer_field.register(
    serializers.QuestionJexlField, lambda field: QuestionJexl
)


ButtonAction = enum_type_from_field(
    "ButtonAction",
    choices=models.Question.ACTION_CHOICES,
    serializer_field=serializers.ButtonActionField,
)

ButtonColor = enum_type_from_field(
    "ButtonColor",
    choices=models.Question.COLOR_CHOICES,
    serializer_field=serializers.ButtonColorField,
)


class Question(Node, graphene.Interface):
    id = graphene.ID(required=True)
    created_at = graphene.DateTime(required=True)
    modified_at = graphene.DateTime(required=True)
    created_by_user = graphene.String()
    created_by_group = graphene.String()
    modified_by_user = graphene.String()
    modified_by_group = graphene.String()
    slug = graphene.String(required=True)
    label = graphene.String(required=True)
    info_text = graphene.String()
    is_required = QuestionJexl(
        required=True,
        description="Required expression is only evaluated when question is not hidden.",
    )
    is_hidden = QuestionJexl(required=True)
    is_archived = graphene.Boolean(required=True)
    meta = generic.GenericScalar(required=True)
    forms = DjangoFilterConnectionField(
        "caluma.caluma_form.schema.Form",
        filterset_class=CollectionFilterSetFactory(
            filterset_class=filters.FormFilterSet, orderset_class=filters.FormOrderSet
        ),
    )
    source = graphene.Field("caluma.caluma_form.schema.Question")

    @classmethod
    def get_queryset(cls, queryset, info):
        queryset = super().get_queryset(queryset, info)
        return (
            queryset.select_related("sub_form", "row_form")
            .order_by("-formquestion__sort")
            .distinct()
        )

    @classmethod
    def resolve_type(cls, instance, info):
        return resolve_question(instance)


class Option(FormDjangoObjectType):
    meta = generic.GenericScalar()

    class Meta:
        model = models.Option
        interfaces = (relay.Node,)
        exclude = ("questions",)
        connection_class = CountableConnectionBase

    @classmethod
    def get_queryset(cls, queryset, info):
        queryset = super().get_queryset(queryset, info)
        return queryset.order_by("-questionoption__sort")


class QuestionConnection(CountableConnectionBase):
    class Meta:
        node = Question


class QuestionQuerysetMixin(object):
    """Mixin to combine all different question types into one queryset."""

    @classmethod
    def get_queryset(cls, queryset, info):
        return Question.get_queryset(queryset, info)


class FormatValidator(ObjectType):
    slug = graphene.String(required=True)
    name = graphene.String(required=True)
    regex = graphene.String(required=True)
    error_msg = graphene.String(required=True)


class FormatValidatorConnection(CountableConnectionBase):
    class Meta:
        node = FormatValidator


class TextQuestion(QuestionQuerysetMixin, FormDjangoObjectType):
    min_length = graphene.Int()
    max_length = graphene.Int()
    placeholder = graphene.String()
    hint_text = graphene.String()
    format_validators = ConnectionField(FormatValidatorConnection)
    default_answer = graphene.Field("caluma.caluma_form.schema.StringAnswer")

    def resolve_format_validators(self, info, *args, **kwargs):
        return get_format_validators(include=self.format_validators)

    class Meta:
        model = models.Question
        exclude = (
            "type",
            "configuration",
            "data_source",
            "options",
            "answers",
            "row_form",
            "sub_form",
            "static_content",
            "dynamicoption_set",
            "calc_expression",
            "calc_dependents",
        )
        use_connection = False
        interfaces = (Question, graphene.Node)


class TextareaQuestion(QuestionQuerysetMixin, FormDjangoObjectType):
    min_length = graphene.Int()
    max_length = graphene.Int()
    placeholder = graphene.String()
    hint_text = graphene.String()
    format_validators = ConnectionField(FormatValidatorConnection)
    default_answer = graphene.Field("caluma.caluma_form.schema.StringAnswer")

    def resolve_format_validators(self, info, *args, **kwargs):
        return get_format_validators(include=self.format_validators)

    class Meta:
        model = models.Question
        exclude = (
            "type",
            "configuration",
            "data_source",
            "options",
            "answers",
            "row_form",
            "sub_form",
            "static_content",
            "dynamicoption_set",
            "calc_expression",
            "calc_dependents",
        )
        use_connection = False
        interfaces = (Question, graphene.Node)


class DateQuestion(QuestionQuerysetMixin, FormDjangoObjectType):
    hint_text = graphene.String()
    default_answer = graphene.Field("caluma.caluma_form.schema.DateAnswer")

    class Meta:
        model = models.Question
        exclude = (
            "type",
            "configuration",
            "data_source",
            "options",
            "answers",
            "row_form",
            "sub_form",
            "placeholder",
            "static_content",
            "format_validators",
            "dynamicoption_set",
            "calc_expression",
            "calc_dependents",
        )
        use_connection = False
        interfaces = (Question, graphene.Node)


class ChoiceQuestion(QuestionQuerysetMixin, FormDjangoObjectType):
    options = DjangoFilterConnectionField(
        Option,
        filterset_class=CollectionFilterSetFactory(
            filterset_class=filters.OptionFilterSet,
            orderset_class=filters.OptionOrderSet,
        ),
    )
    hint_text = graphene.String()
    default_answer = graphene.Field("caluma.caluma_form.schema.StringAnswer")

    class Meta:
        model = models.Question
        exclude = (
            "type",
            "configuration",
            "data_source",
            "answers",
            "row_form",
            "sub_form",
            "placeholder",
            "static_content",
            "format_validators",
            "dynamicoption_set",
            "calc_expression",
            "calc_dependents",
        )
        use_connection = False
        interfaces = (Question, graphene.Node)


class MultipleChoiceQuestion(QuestionQuerysetMixin, FormDjangoObjectType):
    options = DjangoFilterConnectionField(
        Option,
        filterset_class=CollectionFilterSetFactory(
            filterset_class=filters.OptionFilterSet,
            orderset_class=filters.OptionOrderSet,
        ),
    )
    hint_text = graphene.String()
    default_answer = graphene.Field("caluma.caluma_form.schema.ListAnswer")

    class Meta:
        model = models.Question
        exclude = (
            "type",
            "configuration",
            "data_source",
            "answers",
            "row_form",
            "sub_form",
            "placeholder",
            "format_validators",
            "dynamicoption_set",
            "calc_expression",
            "calc_dependents",
        )
        use_connection = False
        interfaces = (Question, graphene.Node)


class DynamicQuestion(graphene.Interface):
    options = ConnectionField(
        DataSourceDataConnection,
        context=graphene.JSONString(
            description="JSON object passed as context to the data source"
        ),
    )
    data_source = graphene.String(required=True)
    hint_text = graphene.String()

    def resolve_options(self, info, context=None, *args, **kwargs):
        return get_data_source_data(info.context.user, self.data_source, self, context)


class DynamicChoiceQuestion(QuestionQuerysetMixin, FormDjangoObjectType):
    hint_text = graphene.String()

    class Meta:
        model = models.Question
        exclude = (
            "type",
            "configuration",
            "answers",
            "row_form",
            "sub_form",
            "placeholder",
            "static_content",
            "format_validators",
            "dynamicoption_set",
            "default_answer",
            "calc_expression",
            "calc_dependents",
        )
        use_connection = False
        interfaces = (Question, DynamicQuestion, graphene.Node)


class DynamicMultipleChoiceQuestion(QuestionQuerysetMixin, FormDjangoObjectType):
    hint_text = graphene.String()

    class Meta:
        model = models.Question
        exclude = (
            "type",
            "configuration",
            "answers",
            "row_form",
            "sub_form",
            "placeholder",
            "static_content",
            "format_validators",
            "dynamicoption_set",
            "default_answer",
            "calc_expression",
            "calc_dependents",
        )
        use_connection = False
        interfaces = (Question, DynamicQuestion, graphene.Node)


class DynamicOption(DjangoObjectType):
    question = graphene.Field(
        "caluma.caluma_form.schema.DynamicQuestion", required=True
    )

    class Meta:
        model = models.DynamicOption
        interfaces = (relay.Node,)
        connection_class = CountableConnectionBase
        fields = "__all__"


class IntegerQuestion(QuestionQuerysetMixin, FormDjangoObjectType):
    max_value = graphene.Int()
    min_value = graphene.Int()
    placeholder = graphene.String()
    hint_text = graphene.String()
    default_answer = graphene.Field("caluma.caluma_form.schema.IntegerAnswer")

    class Meta:
        model = models.Question
        exclude = (
            "type",
            "configuration",
            "data_source",
            "options",
            "answers",
            "row_form",
            "sub_form",
            "static_content",
            "format_validators",
            "dynamicoption_set",
            "calc_expression",
            "calc_dependents",
        )
        use_connection = False
        interfaces = (Question, graphene.Node)


class FloatQuestion(QuestionQuerysetMixin, FormDjangoObjectType):
    min_value = graphene.Float()
    max_value = graphene.Float()
    placeholder = graphene.String()
    hint_text = graphene.String()
    default_answer = graphene.Field("caluma.caluma_form.schema.FloatAnswer")

    class Meta:
        model = models.Question
        exclude = (
            "type",
            "configuration",
            "data_source",
            "options",
            "answers",
            "row_form",
            "sub_form",
            "static_content",
            "format_validators",
            "dynamicoption_set",
            "calc_expression",
            "calc_dependents",
        )
        use_connection = False
        interfaces = (Question, graphene.Node)


class TableQuestion(QuestionQuerysetMixin, FormDjangoObjectType):
    hint_text = graphene.String()
    default_answer = graphene.Field("caluma.caluma_form.schema.TableAnswer")

    class Meta:
        model = models.Question
        exclude = (
            "type",
            "configuration",
            "data_source",
            "options",
            "answers",
            "sub_form",
            "placeholder",
            "static_content",
            "format_validators",
            "dynamicoption_set",
            "calc_expression",
            "calc_dependents",
        )
        use_connection = False
        interfaces = (Question, graphene.Node)


class FormQuestion(QuestionQuerysetMixin, FormDjangoObjectType):
    class Meta:
        model = models.Question
        exclude = (
            "type",
            "configuration",
            "data_source",
            "options",
            "answers",
            "row_form",
            "placeholder",
            "hint_text",
            "static_content",
            "format_validators",
            "dynamicoption_set",
            "default_answer",
            "calc_expression",
            "calc_dependents",
        )
        use_connection = False
        interfaces = (Question, graphene.Node)


class FilesQuestion(QuestionQuerysetMixin, FormDjangoObjectType):
    hint_text = graphene.String()

    class Meta:
        model = models.Question
        exclude = (
            "type",
            "configuration",
            "data_source",
            "options",
            "answers",
            "row_form",
            "sub_form",
            "placeholder",
            "static_content",
            "format_validators",
            "dynamicoption_set",
            "default_answer",
            "calc_expression",
            "calc_dependents",
        )
        use_connection = False
        interfaces = (Question, graphene.Node)


class StaticQuestion(QuestionQuerysetMixin, FormDjangoObjectType):
    is_required = QuestionJexl(
        required=True,
        description="Required expression is only evaluated when question is not hidden."
        " This should not be used for `StaticQuestion`, because it can never be satisfied.",
    )

    class Meta:
        model = models.Question
        exclude = (
            "type",
            "configuration",
            "options",
            "answers",
            "row_form",
            "sub_form",
            "placeholder",
            "hint_text",
            "format_validators",
            "dynamicoption_set",
            "default_answer",
            "calc_expression",
            "calc_dependents",
        )
        use_connection = False
        interfaces = (Question, graphene.Node)


class CalculatedFloatQuestion(QuestionQuerysetMixin, FormDjangoObjectType):
    hint_text = graphene.String()

    class Meta:
        model = models.Question
        exclude = (
            "type",
            "configuration",
            "data_source",
            "options",
            "answers",
            "row_form",
            "sub_form",
            "placeholder",
            "static_content",
            "format_validators",
            "dynamicoption_set",
            "default_answer",
            "calc_dependents",
        )
        use_connection = False
        interfaces = (Question, graphene.Node)


class ActionButtonQuestion(QuestionQuerysetMixin, FormDjangoObjectType):
    action = ButtonAction(required=True)
    color = ButtonColor(required=True)
    validate_on_enter = graphene.Boolean(required=True)

    class Meta:
        model = models.Question
        exclude = (
            "type",
            "configuration",
            "data_source",
            "options",
            "answers",
            "row_form",
            "sub_form",
            "placeholder",
            "hint_text",
            "static_content",
            "format_validators",
            "dynamicoption_set",
            "default_answer",
            "calc_expression",
            "calc_dependents",
        )
        use_connection = False
        interfaces = (Question, graphene.Node)


class Form(FormDjangoObjectType):
    questions = DjangoFilterInterfaceConnectionField(
        QuestionConnection,
        filterset_class=CollectionFilterSetFactory(
            filterset_class=filters.QuestionFilterSet,
            orderset_class=filters.QuestionOrderSet,
        ),
    )
    meta = generic.GenericScalar()

    class Meta:
        model = models.Form
        interfaces = (relay.Node,)
        exclude = ("workflows", "tasks")
        connection_class = CountableConnectionBase


class SaveForm(UserDefinedPrimaryKeyMixin, Mutation):
    class Meta:
        serializer_class = serializers.SaveFormSerializer


class CopyForm(UserDefinedPrimaryKeyMixin, Mutation):
    class Meta:
        serializer_class = serializers.CopyFormSerializer
        model_operations = ["create"]


class AddFormQuestion(Mutation):
    """Add question at the end of form."""

    class Meta:
        lookup_input_kwarg = "form"
        serializer_class = serializers.AddFormQuestionSerializer


class RemoveFormQuestion(Mutation):
    class Meta:
        lookup_input_kwarg = "form"
        serializer_class = serializers.RemoveFormQuestionSerializer


class ReorderFormQuestions(Mutation):
    class Meta:
        lookup_input_kwarg = "form"
        serializer_class = serializers.ReorderFormQuestionsSerializer


class CopyQuestion(UserDefinedPrimaryKeyMixin, Mutation):
    class Meta:
        serializer_class = serializers.CopyQuestionSerializer
        return_field_type = Question
        model_operations = ["create"]


class SaveQuestion(UserDefinedPrimaryKeyMixin, Mutation):
    """
    Base class of all save question mutations.

    Defined so it is easy to set a permission for all different types
    of questions.

    See `caluma.permissions.BasePermission` for more details.
    """

    class Meta:
        abstract = True


class SaveTextQuestion(SaveQuestion):
    class Meta:
        serializer_class = serializers.SaveTextQuestionSerializer
        return_field_type = Question


class SaveTextareaQuestion(SaveQuestion):
    class Meta:
        serializer_class = serializers.SaveTextareaQuestionSerializer
        return_field_type = Question


class SaveDateQuestion(SaveQuestion):
    class Meta:
        serializer_class = serializers.SaveDateQuestionSerializer
        return_field_type = Question


class SaveChoiceQuestion(SaveQuestion):
    class Meta:
        serializer_class = serializers.SaveChoiceQuestionSerializer
        return_field_type = Question


class SaveMultipleChoiceQuestion(SaveQuestion):
    class Meta:
        serializer_class = serializers.SaveMultipleChoiceQuestionSerializer
        return_field_type = Question


class SaveDynamicChoiceQuestion(SaveQuestion):
    class Meta:
        serializer_class = serializers.SaveDynamicChoiceQuestionSerializer
        return_field_type = Question


class SaveDynamicMultipleChoiceQuestion(SaveQuestion):
    class Meta:
        serializer_class = serializers.SaveDynamicMultipleChoiceQuestionSerializer
        return_field_type = Question


class SaveIntegerQuestion(SaveQuestion):
    class Meta:
        serializer_class = serializers.SaveIntegerQuestionSerializer
        return_field_type = Question


class SaveFloatQuestion(SaveQuestion):
    class Meta:
        serializer_class = serializers.SaveFloatQuestionSerializer
        return_field_type = Question


class SaveTableQuestion(SaveQuestion):
    class Meta:
        serializer_class = serializers.SaveTableQuestionSerializer
        return_field_type = Question


class SaveFormQuestion(SaveQuestion):
    class Meta:
        serializer_class = serializers.SaveFormQuestionSerializer
        return_field_type = Question


class SaveFilesQuestion(SaveQuestion):
    class Meta:
        serializer_class = serializers.SaveFileQuestionSerializer
        return_field_type = Question


class SaveStaticQuestion(SaveQuestion):
    class Meta:
        serializer_class = serializers.SaveStaticQuestionSerializer
        return_field_type = Question


class SaveCalculatedFloatQuestion(SaveQuestion):
    class Meta:
        serializer_class = serializers.SaveCalculatedFloatQuestionSerializer
        return_field_type = Question


class SaveActionButtonQuestion(SaveQuestion):
    class Meta:
        serializer_class = serializers.SaveActionButtonQuestionSerializer
        return_field_type = Question


class SaveOption(UserDefinedPrimaryKeyMixin, Mutation):
    class Meta:
        serializer_class = serializers.SaveOptionSerializer


class CopyOption(UserDefinedPrimaryKeyMixin, Mutation):
    class Meta:
        serializer_class = serializers.CopyOptionSerializer
        model_operations = ["create"]


class Answer(Node, graphene.Interface):
    id = graphene.ID()
    created_at = graphene.DateTime(required=True)
    modified_at = graphene.DateTime(required=True)
    created_by_user = graphene.String()
    created_by_group = graphene.String()
    modified_by_user = graphene.String()
    modified_by_group = graphene.String()
    question = graphene.Field(Question, required=True)
    meta = generic.GenericScalar(required=True)

    @classmethod
    def resolve_type(cls, instance, info):
        return resolve_answer(instance)


class AnswerQuerysetMixin(object):
    """Mixin to combine all different answer types into one queryset."""

    @classmethod
    def get_queryset(cls, queryset, info):
        return Answer.get_queryset(queryset, info)


class IntegerAnswer(AnswerQuerysetMixin, FormDjangoObjectType):
    value = graphene.Int()

    class Meta:
        model = models.Answer
        exclude = ("document", "documents", "files", "date")
        use_connection = False
        interfaces = (Answer, graphene.Node)


class FloatAnswer(AnswerQuerysetMixin, FormDjangoObjectType):
    value = graphene.Float()

    class Meta:
        model = models.Answer
        exclude = ("document", "documents", "files", "date")
        use_connection = False
        interfaces = (Answer, graphene.Node)


class DateAnswer(AnswerQuerysetMixin, FormDjangoObjectType):
    value = graphene.types.datetime.Date()

    def resolve_value(self, info, **args):
        return self.date

    class Meta:
        model = models.Answer
        exclude = ("document", "documents", "files")
        use_connection = False
        interfaces = (Answer, graphene.Node)


class SelectedOption(ObjectType):
    label = graphene.String(required=True)
    slug = graphene.String(required=True)


class SelectedOptionConnection(CountableConnectionBase):
    class Meta:
        node = SelectedOption


class StringAnswer(AnswerQuerysetMixin, FormDjangoObjectType):
    value = graphene.String()
    selected_option = graphene.Field(SelectedOption)

    def resolve_selected_option(self, info, **args):
        selected_options = self.selected_options
        return selected_options[0] if selected_options else None

    class Meta:
        model = models.Answer
        exclude = ("document", "documents", "files", "date")
        use_connection = False
        interfaces = (Answer, graphene.Node)


class ListAnswer(AnswerQuerysetMixin, FormDjangoObjectType):
    value = graphene.List(graphene.String)
    selected_options = ConnectionField(SelectedOptionConnection)

    class Meta:
        model = models.Answer
        exclude = ("document", "documents", "files", "date")
        use_connection = False
        interfaces = (Answer, graphene.Node)


class AnswerConnection(CountableConnectionBase):
    class Meta:
        node = Answer


class Document(FormDjangoObjectType):
    answers = DjangoFilterInterfaceConnectionField(
        AnswerConnection,
        filterset_class=CollectionFilterSetFactory(
            filterset_class=filters.AnswerFilterSet,
            orderset_class=filters.AnswerOrderSet,
        ),
    )
    meta = generic.GenericScalar()
    modified_content_at = graphene.DateTime()
    modified_content_by_user = graphene.String()
    modified_content_by_group = graphene.String()

    class Meta:
        model = models.Document
        exclude = ("family", "dynamicoption_set")
        interfaces = (graphene.Node,)
        connection_class = CountableConnectionBase


class TableAnswer(AnswerQuerysetMixin, FormDjangoObjectType):
    value = graphene.List(Document)

    def resolve_value(self, info, **args):
        return self.documents.order_by("-answerdocument__sort")

    class Meta:
        model = models.Answer
        exclude = ("documents", "files", "date")
        use_connection = False
        interfaces = (Answer, graphene.Node)


class File(FormDjangoObjectType):
    name = graphene.String(required=True)
    upload_url = graphene.String()
    download_url = graphene.String()
    metadata = generic.GenericScalar()

    class Meta:
        model = models.File
        interfaces = (relay.Node,)
        fields = "__all__"


class FilesAnswer(AnswerQuerysetMixin, FormDjangoObjectType):
    value = graphene.List(File, required=True)

    def resolve_value(self, info, **args):
        return self.files.all()

    class Meta:
        model = models.Answer
        exclude = ("document", "documents", "date", "files")
        use_connection = False
        interfaces = (Answer, graphene.Node)


class SaveDocument(Mutation):
    class Input:
        id = graphene.String()

    class Meta:
        serializer_class = serializers.DocumentSerializer
        model_operations = ["create", "update"]


class CopyDocument(Mutation):
    class Meta:
        serializer_class = serializers.CopyDocumentSerializer
        model_operations = ["create"]


class SaveDocumentAnswer(Mutation):
    @classmethod
    def get_object(cls, root, info, queryset, **input):
        question_id = extract_global_id(input["question"])
        document_id = extract_global_id(input["document"])
        instance = models.Answer.objects.filter(
            question=question_id, document=document_id
        ).first()
        return instance

    class Meta:
        abstract = True


class SaveDocumentStringAnswer(SaveDocumentAnswer):
    class Meta:
        serializer_class = serializers.SaveDocumentStringAnswerSerializer
        return_field_type = Answer


class SaveDocumentListAnswer(SaveDocumentAnswer):
    class Meta:
        serializer_class = serializers.SaveDocumentListAnswerSerializer
        return_field_type = Answer


class SaveDocumentIntegerAnswer(SaveDocumentAnswer):
    class Meta:
        serializer_class = serializers.SaveDocumentIntegerAnswerSerializer
        return_field_type = Answer


class SaveDocumentFloatAnswer(SaveDocumentAnswer):
    class Meta:
        serializer_class = serializers.SaveDocumentFloatAnswerSerializer
        return_field_type = Answer


class SaveDocumentDateAnswer(SaveDocumentAnswer):
    class Meta:
        serializer_class = serializers.SaveDocumentDateAnswerSerializer
        return_field_type = Answer


class SaveDocumentTableAnswer(SaveDocumentAnswer):
    class Meta:
        serializer_class = serializers.SaveDocumentTableAnswerSerializer
        return_field_type = Answer


class SaveFile(graphene.InputObjectType):
    id = graphene.String()
    name = graphene.String()


class SaveDocumentFilesAnswer(SaveDocumentAnswer):
    class Input:
        value = graphene.List(SaveFile, required=False)

    class Meta:
        serializer_class = serializers.SaveDocumentFilesAnswerSerializer
        return_field_type = Answer


class SaveDefaultAnswer(Mutation):
    @classmethod
    def get_object(cls, root, info, queryset, **input):
        question_id = extract_global_id(input["question"])
        instance = models.Question.objects.get(pk=question_id).default_answer
        return instance

    class Meta:
        abstract = True


class SaveDefaultStringAnswer(SaveDefaultAnswer):
    class Meta:
        serializer_class = serializers.SaveDefaultStringAnswerSerializer
        return_field_type = Answer


class SaveDefaultListAnswer(SaveDefaultAnswer):
    class Meta:
        serializer_class = serializers.SaveDefaultListAnswerSerializer
        return_field_type = Answer


class SaveDefaultIntegerAnswer(SaveDefaultAnswer):
    class Meta:
        serializer_class = serializers.SaveDefaultIntegerAnswerSerializer
        return_field_type = Answer


class SaveDefaultFloatAnswer(SaveDefaultAnswer):
    class Meta:
        serializer_class = serializers.SaveDefaultFloatAnswerSerializer
        return_field_type = Answer


class SaveDefaultDateAnswer(SaveDefaultAnswer):
    class Meta:
        serializer_class = serializers.SaveDefaultDateAnswerSerializer
        return_field_type = Answer


class SaveDefaultTableAnswer(SaveDefaultAnswer):
    class Meta:
        serializer_class = serializers.SaveDefaultTableAnswerSerializer
        return_field_type = Answer


class RemoveAnswer(Mutation):
    class Meta:
        lookup_input_kwarg = "answer"
        serializer_class = serializers.RemoveAnswerSerializer
        return_field_type = Answer


class RemoveDefaultAnswer(Mutation):
    class Meta:
        lookup_input_kwarg = "question"
        serializer_class = serializers.RemoveDefaultAnswerSerializer
        return_field_type = Question


class RemoveDocument(Mutation):
    class Meta:
        lookup_input_kwarg = "document"
        serializer_class = serializers.RemoveDocumentSerializer


class Mutation(object):
    save_form = SaveForm().Field()
    copy_form = CopyForm().Field()
    add_form_question = AddFormQuestion().Field()
    remove_form_question = RemoveFormQuestion().Field()
    reorder_form_questions = ReorderFormQuestions().Field()

    save_option = SaveOption().Field()
    copy_option = CopyOption().Field()

    copy_question = CopyQuestion().Field()
    save_text_question = SaveTextQuestion().Field()
    save_textarea_question = SaveTextareaQuestion().Field()
    save_date_question = SaveDateQuestion().Field()
    save_choice_question = SaveChoiceQuestion().Field()
    save_multiple_choice_question = SaveMultipleChoiceQuestion().Field()
    save_dynamic_choice_question = SaveDynamicChoiceQuestion().Field()
    save_dynamic_multiple_choice_question = SaveDynamicMultipleChoiceQuestion().Field()
    save_float_question = SaveFloatQuestion().Field()
    save_integer_question = SaveIntegerQuestion().Field()
    save_table_question = SaveTableQuestion().Field()
    save_form_question = SaveFormQuestion().Field()
    save_files_question = SaveFilesQuestion().Field()
    save_static_question = SaveStaticQuestion().Field()
    save_calculated_float_question = SaveCalculatedFloatQuestion().Field()
    save_action_button_question = SaveActionButtonQuestion().Field()

    copy_document = CopyDocument().Field()
    save_document = SaveDocument().Field()
    save_document_string_answer = SaveDocumentStringAnswer().Field()
    save_document_integer_answer = SaveDocumentIntegerAnswer().Field()
    save_document_float_answer = SaveDocumentFloatAnswer().Field()
    save_document_date_answer = SaveDocumentDateAnswer().Field()
    save_document_list_answer = SaveDocumentListAnswer().Field()
    save_document_table_answer = SaveDocumentTableAnswer().Field()
    save_document_files_answer = SaveDocumentFilesAnswer().Field()

    save_default_string_answer = SaveDefaultStringAnswer().Field()
    save_default_integer_answer = SaveDefaultIntegerAnswer().Field()
    save_default_float_answer = SaveDefaultFloatAnswer().Field()
    save_default_date_answer = SaveDefaultDateAnswer().Field()
    save_default_list_answer = SaveDefaultListAnswer().Field()
    save_default_table_answer = SaveDefaultTableAnswer().Field()

    remove_answer = RemoveAnswer().Field()
    remove_default_answer = RemoveDefaultAnswer().Field()
    remove_document = RemoveDocument().Field()


class ValidationEntry(ObjectType):
    slug = graphene.String(required=True)
    error_msg = graphene.String(required=True)


class ValidationResult(ObjectType):
    id = graphene.ID()
    id.__doc__ = "References the document ID"

    is_valid = graphene.Boolean()
    errors = graphene.List(ValidationEntry)


class DocumentValidityConnection(CountableConnectionBase):
    class Meta:
        node = ValidationResult


def validate_document(info, document_global_id):

    document_id = extract_global_id(document_global_id)

    document_qs = Document.get_queryset(models.Document.objects.all(), info)

    document = get_object_or_404(document_qs, pk=document_id)
    result = get_document_validity(document, info)

    errors = result.pop("errors")
    result = ValidationResult(
        **result, errors=[ValidationEntry(**err) for err in errors]
    )

    return [result]


class Query:
    all_forms = DjangoFilterConnectionField(
        Form,
        filterset_class=CollectionFilterSetFactory(
            filterset_class=filters.FormFilterSet,
            orderset_class=filters.FormOrderSet,
        ),
    )
    all_questions = DjangoFilterInterfaceConnectionField(
        QuestionConnection,
        filterset_class=CollectionFilterSetFactory(
            filterset_class=filters.QuestionFilterSet,
            orderset_class=filters.QuestionOrderSet,
        ),
    )
    all_documents = DjangoFilterConnectionField(
        Document,
        filterset_class=CollectionFilterSetFactory(
            filterset_class=filters.DocumentFilterSet,
            orderset_class=filters.DocumentOrderSet,
        ),
    )
    all_format_validators = ConnectionField(FormatValidatorConnection)
    all_used_dynamic_options = DjangoFilterConnectionField(
        DynamicOption,
        filterset_class=CollectionFilterSetFactory(
            filterset_class=filters.DynamicOptionFilterSet,
            orderset_class=filters.DynamicOptionOrderSet,
        ),
    )

    document_validity = ConnectionField(
        DocumentValidityConnection, id=graphene.ID(required=True)
    )

    def resolve_all_format_validators(self, info, **kwargs):
        return get_format_validators()

    def resolve_document_validity(self, info, id, **kwargs):
        return validate_document(info, id)


QUESTION_ANSWER_TYPES = {
    models.Question.TYPE_MULTIPLE_CHOICE: ListAnswer,
    models.Question.TYPE_INTEGER: IntegerAnswer,
    models.Question.TYPE_FLOAT: FloatAnswer,
    models.Question.TYPE_DATE: DateAnswer,
    models.Question.TYPE_CHOICE: StringAnswer,
    models.Question.TYPE_TEXTAREA: StringAnswer,
    models.Question.TYPE_TEXT: StringAnswer,
    models.Question.TYPE_TABLE: TableAnswer,
    models.Question.TYPE_FILES: FilesAnswer,
    models.Question.TYPE_DYNAMIC_CHOICE: StringAnswer,
    models.Question.TYPE_DYNAMIC_MULTIPLE_CHOICE: ListAnswer,
    models.Question.TYPE_CALCULATED_FLOAT: FloatAnswer,
}

QUESTION_OBJECT_TYPES = {
    models.Question.TYPE_TEXT: TextQuestion,
    models.Question.TYPE_FLOAT: FloatQuestion,
    models.Question.TYPE_CHOICE: ChoiceQuestion,
    models.Question.TYPE_INTEGER: IntegerQuestion,
    models.Question.TYPE_MULTIPLE_CHOICE: MultipleChoiceQuestion,
    models.Question.TYPE_DYNAMIC_CHOICE: DynamicChoiceQuestion,
    models.Question.TYPE_DYNAMIC_MULTIPLE_CHOICE: DynamicMultipleChoiceQuestion,
    models.Question.TYPE_TEXTAREA: TextareaQuestion,
    models.Question.TYPE_DATE: DateQuestion,
    models.Question.TYPE_TABLE: TableQuestion,
    models.Question.TYPE_FORM: FormQuestion,
    models.Question.TYPE_FILES: FilesQuestion,
    models.Question.TYPE_STATIC: StaticQuestion,
    models.Question.TYPE_CALCULATED_FLOAT: CalculatedFloatQuestion,
    models.Question.TYPE_ACTION_BUTTON: ActionButtonQuestion,
}



from datetime import timedelta

from django.contrib.postgres.fields import ArrayField
from django.contrib.postgres.indexes import GinIndex
from django.db import models
from django.db.models.signals import post_init, pre_save
from django.dispatch import receiver
from django.utils import timezone
from localized_fields.fields import LocalizedField

from ..caluma_core.models import SlugModel, UUIDModel


class Task(SlugModel):
    TYPE_SIMPLE = "simple"
    TYPE_COMPLETE_WORKFLOW_FORM = "complete_workflow_form"
    TYPE_COMPLETE_TASK_FORM = "complete_task_form"

    TYPE_CHOICES = (TYPE_SIMPLE, TYPE_COMPLETE_WORKFLOW_FORM, TYPE_COMPLETE_TASK_FORM)
    TYPE_CHOICES_TUPLE = (
        (TYPE_SIMPLE, "Task which can simply be marked as completed."),
        (TYPE_COMPLETE_WORKFLOW_FORM, "Task to complete a defined workflow form."),
        (TYPE_COMPLETE_TASK_FORM, "Task to complete a defined task form."),
    )

    name = LocalizedField(blank=False, null=False, required=False)
    description = LocalizedField(blank=True, null=True, required=False)
    type = models.CharField(choices=TYPE_CHOICES_TUPLE, max_length=50)
    meta = models.JSONField(default=dict)
    address_groups = models.TextField(
        blank=True,
        null=True,
        help_text="Group jexl returning what group(s) derived work items will be addressed to.",
    )
    control_groups = models.TextField(
        blank=True,
        null=True,
        help_text="Group jexl returning what group(s) derived work items will be assigned to for controlling.",
    )
    is_archived = models.BooleanField(default=False)
    form = models.ForeignKey(
        "caluma_form.Form",
        on_delete=models.DO_NOTHING,
        related_name="tasks",
        blank=True,
        null=True,
    )
    lead_time = models.PositiveIntegerField(
        blank=True,
        null=True,
        help_text="Time in seconds task may take to be processed.",
    )
    is_multiple_instance = models.BooleanField(
        default=False,
        help_text="Allows creating multiple work items for this task using the `CreateWorkItem` mutation. If true, one work item will be created for each entry in `address_groups`.",
    )
    continue_async = models.BooleanField(
        default=False,
        help_text="Whether to continue the flow if the multiple instance work item has ready siblings",
    )

    def calculate_deadline(self):
        if self.lead_time is not None:
            return timezone.now() + timedelta(seconds=self.lead_time)
        return None

    class Meta:
        indexes = [GinIndex(fields=["meta"])]


class Workflow(SlugModel):
    name = LocalizedField(blank=False, null=False, required=False)
    description = LocalizedField(blank=True, null=True, required=False)
    meta = models.JSONField(default=dict)
    is_published = models.BooleanField(default=False)
    is_archived = models.BooleanField(default=False)
    start_tasks = models.ManyToManyField(
        Task, related_name="+", help_text="Starting task(s) of the workflow."
    )
    allow_all_forms = models.BooleanField(
        default=False, help_text="Allow workflow to be started with any form"
    )
    allow_forms = models.ManyToManyField(
        "caluma_form.Form",
        help_text="List of forms which are allowed to start workflow with",
        related_name="workflows",
        blank=True,
    )

    @property
    def flows(self):
        return Flow.objects.filter(pk__in=self.task_flows.values("flow"))

    class Meta:
        indexes = [GinIndex(fields=["meta"])]


class Flow(UUIDModel):
    next = models.TextField()


class TaskFlow(UUIDModel):
    workflow = models.ForeignKey(
        Workflow, on_delete=models.CASCADE, related_name="task_flows"
    )
    task = models.ForeignKey(Task, related_name="task_flows", on_delete=models.CASCADE)
    flow = models.ForeignKey(
        Flow, related_name="task_flows", on_delete=models.CASCADE, null=True, blank=True
    )
    redoable = models.TextField(
        blank=True,
        null=True,
        help_text="jexl returning what tasks can be redone from this taskflow.",
    )

    class Meta:
        unique_together = ("workflow", "task")


class Case(UUIDModel):
    STATUS_RUNNING = "running"
    STATUS_COMPLETED = "completed"
    STATUS_CANCELED = "canceled"
    STATUS_SUSPENDED = "suspended"

    STATUS_CHOICE_TUPLE = (
        (STATUS_RUNNING, "Case is running and work items need to be completed."),
        (STATUS_COMPLETED, "Case is done."),
        (STATUS_CANCELED, "Case is canceled."),
        (STATUS_SUSPENDED, "Case is suspended."),
    )

    family = models.ForeignKey(
        "self",
        help_text="Family id which case belongs to.",
        null=True,
        on_delete=models.CASCADE,
        related_name="+",
    )

    closed_at = models.DateTimeField(
        blank=True,
        null=True,
        help_text="Time when case has either been canceled or completed",
    )
    closed_by_user = models.CharField(max_length=150, blank=True, null=True)
    closed_by_group = models.CharField(max_length=150, blank=True, null=True)

    workflow = models.ForeignKey(
        Workflow, related_name="cases", on_delete=models.DO_NOTHING
    )
    status = models.CharField(choices=STATUS_CHOICE_TUPLE, max_length=50, db_index=True)
    meta = models.JSONField(default=dict)
    document = models.OneToOneField(
        "caluma_form.Document",
        on_delete=models.PROTECT,
        related_name="case",
        blank=True,
        null=True,
    )

    class Meta:
        indexes = [GinIndex(fields=["meta"])]


@receiver(post_init, sender=Case)
def set_case_family(sender, instance, **kwargs):
    """
    Ensure case has the family pointer set.

    This sets the case's family if not overruled or set already.
    """
    if instance.family_id is None:
        instance.family = instance


class WorkItem(UUIDModel):
    STATUS_READY = "ready"
    STATUS_COMPLETED = "completed"
    STATUS_CANCELED = "canceled"
    STATUS_SKIPPED = "skipped"
    STATUS_SUSPENDED = "suspended"
    STATUS_REDO = "redo"

    STATUS_CHOICE_TUPLE = (
        (STATUS_READY, "Work item is ready to be processed."),
        (STATUS_COMPLETED, "Work item is done."),
        (STATUS_CANCELED, "Work item is canceled."),
        (STATUS_SKIPPED, "Work item is skipped."),
        (STATUS_SUSPENDED, "Work item is suspended."),
        (STATUS_REDO, "Work item has been marked for redo."),
    )

    name = LocalizedField(
        blank=False,
        null=False,
        required=False,
        help_text="Will be set from Task, if not provided.",
    )
    description = LocalizedField(
        blank=True,
        null=True,
        required=False,
        help_text="Will be set from Task, if not provided.",
    )

    closed_at = models.DateTimeField(
        blank=True,
        null=True,
        help_text="Time when work item has either been canceled or completed",
    )
    closed_by_user = models.CharField(max_length=150, blank=True, null=True)
    closed_by_group = models.CharField(max_length=150, blank=True, null=True)
    deadline = models.DateTimeField(blank=True, null=True)

    task = models.ForeignKey(
        Task, on_delete=models.DO_NOTHING, related_name="work_items"
    )
    status = models.CharField(choices=STATUS_CHOICE_TUPLE, max_length=50, db_index=True)
    meta = models.JSONField(default=dict)

    addressed_groups = ArrayField(
        models.CharField(max_length=150),
        default=list,
        help_text=(
            "Offer work item to be processed by a group of users, "
            "such are not committed to process it though."
        ),
    )

    controlling_groups = ArrayField(
        models.CharField(max_length=150),
        default=list,
        help_text="List of groups this work item is assigned to for controlling.",
    )

    assigned_users = ArrayField(
        models.CharField(max_length=150),
        default=list,
        help_text="Users responsible to undertake given work item.",
    )

    case = models.ForeignKey(Case, related_name="work_items", on_delete=models.CASCADE)
    child_case = models.OneToOneField(
        Case,
        related_name="parent_work_item",
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        help_text="Defines case of a sub-workflow",
    )

    document = models.OneToOneField(
        "caluma_form.Document",
        on_delete=models.PROTECT,
        related_name="work_item",
        blank=True,
        null=True,
    )

    previous_work_item = models.ForeignKey(
        "self",
        on_delete=models.SET_NULL,
        related_name="succeeding_work_items",
        blank=True,
        null=True,
    )

    def get_redoable(self):
        """
        Get redoable tasks for this WorkItem.

        For this we evaluate the `redoable`-field of the WorkItem-tasks TaskFlow.

        :return: QuerySet
        """
        task_flow = self.task.task_flows.filter(workflow=self.case.workflow).first()

        if not task_flow:
            return Task.objects.none()

        jexl = task_flow.redoable

        if not jexl:
            return Task.objects.none()

        from .jexl import FlowJexl

        slugs = FlowJexl(case=self.case, prev_work_item=self).evaluate(jexl)

        if not isinstance(slugs, list):
            slugs = [slugs]

        return Task.objects.filter(pk__in=slugs)

    class Meta:
        indexes = [
            GinIndex(fields=["addressed_groups"]),
            GinIndex(fields=["assigned_users"]),
            GinIndex(fields=["meta"]),
        ]


@receiver(pre_save, sender=WorkItem)
def set_name_and_description(sender, instance, **kwargs):
    """
    Ensure WorkItem has a name and description set.

    Default to values from Task.
    """
    if not any(instance.name.values()):
        instance.name = instance.task.name
    if not any(instance.description.values()):
        instance.description = instance.task.description



import graphene
from graphene import ConnectionField, String, relay
from graphene.types import ObjectType, generic

from ..caluma_core.filters import (
    CollectionFilterSetFactory,
    DjangoFilterConnectionField,
)
from ..caluma_core.mutation import Mutation, UserDefinedPrimaryKeyMixin
from ..caluma_core.types import (
    CountableConnectionBase,
    DjangoObjectType,
    enum_type_from_field,
)
from . import filters, models, serializers

AggregateFunction = enum_type_from_field(
    "AggregateFunction",
    models.AnalyticsField.function,
    description="Aggregate function for pivot table",
    serializer_field=serializers.AggregateFunctionField,
)


class AvailableField(ObjectType):
    """Available fields show users what can be selected in an analysis.

    The main identifier is the source path, but for display purposes,
    a label (field at current position) and a full_label (including
    parent fields' labels) is available.

    Frontends should query sub-fields (via prefix/depth) if is_leaf is
    False. Some fields can be non-leafs as well as values, such as
    dates: Dates can be extracted "as is", or we can extract a
    date part (such as year, quarter, ...) from it.
    """

    label = String()
    full_label = String()
    source_path = String()
    is_leaf = graphene.Boolean()
    is_value = graphene.Boolean()
    supported_functions = graphene.List(AggregateFunction)

    class Meta:
        interfaces = (relay.Node,)
        connection_class = CountableConnectionBase


class AvailableFieldConnection(CountableConnectionBase):
    class Meta:
        node = AvailableField


class AnalyticsCell(ObjectType):
    """A cell represents one value in the analytics output."""

    alias = String()
    value = String(required=False)


class AnalyticsRow(graphene.Connection):
    class Meta:
        node = AnalyticsCell


class AnalyticsTableContent(graphene.Connection):
    class Meta:
        node = AnalyticsRow


class AnalyticsOutput(ObjectType):
    records = ConnectionField(AnalyticsTableContent)
    summary = graphene.Field(AnalyticsRow)

    @staticmethod
    def resolve_records(table, info, *args, **kwargs):
        rows = [
            AnalyticsRow(
                edges=[
                    {"node": {"alias": alias, "value": row[alias]}}
                    for alias in table.field_ordering
                ]
            )
            for row in table.get_records()
        ]
        return rows

    @staticmethod
    def resolve_summary(table, info, *args, **kwargs):
        summary_row = table.get_summary()
        if summary_row:
            return AnalyticsRow(
                edges=[
                    {"node": {"alias": alias, "value": summary_row[alias]}}
                    for alias in table.field_ordering
                ]
            )
        return AnalyticsRow(edges=[])


StartingObject = enum_type_from_field(
    "StartingObject",
    models.AnalyticsTable.starting_object,
    serializer_field=serializers.StartingObjectField,
)


class AnalyticsTable(DjangoObjectType):
    available_fields = ConnectionField(
        AvailableFieldConnection,
        prefix=String(required=False),
        depth=graphene.Int(required=False),
    )
    result_data = graphene.Field(AnalyticsOutput)
    starting_object = StartingObject(required=False)

    @staticmethod
    def resolve_available_fields(obj, info, prefix=None, depth=None, **kwargs):
        start = obj.get_starting_object(info)

        depth = depth if depth and depth > 0 else 1
        prefix = prefix.split(".") if prefix else []

        return sorted(
            [
                {
                    "id": ".".join(field.source_path()),
                    "label": field.label,
                    "full_label": field.full_label(),
                    "source_path": ".".join(field.source_path()),
                    "is_leaf": field.is_leaf(),
                    "is_value": field.is_value(),
                    "supported_functions": field.supported_functions(),
                }
                for path, field in start.get_fields(prefix, depth).items()
            ],
            key=lambda field: field["id"],
        )

    @staticmethod
    def resolve_result_data(obj, info):
        return obj.get_analytics(info)

    class Meta:
        model = models.AnalyticsTable
        interfaces = (relay.Node,)
        connection_class = CountableConnectionBase
        fields = [
            "slug",
            "meta",
            "created_by_group",
            "created_by_user",
            "modified_by_group",
            "modified_by_user",
            "created_at",
            "modified_at",
            "disable_visibilities",
            "available_fields",
            "result_data",
            "fields",
            "name",
            "starting_object",
        ]


class AnalyticsField(DjangoObjectType):
    meta = generic.GenericScalar()
    filters = graphene.List(String, required=False)
    function = AggregateFunction(required=False)

    class Meta:
        model = models.AnalyticsField
        interfaces = (relay.Node,)
        connection_class = CountableConnectionBase
        fields = "__all__"


class SaveAnalyticsTable(UserDefinedPrimaryKeyMixin, Mutation):
    class Meta:
        serializer_class = serializers.SaveAnalyticsTableSerializer
        model_operations = ["create", "update"]


class RemoveAnalyticsTable(Mutation):
    class Meta:
        serializer_class = serializers.RemoveAnalyticsTableSerializer
        lookup_input_kwarg = "slug"
        model_operations = ["update"]


class SaveAnalyticsField(Mutation):
    class Meta:
        serializer_class = serializers.SaveAnalyticsFieldSerializer
        model_operations = ["create", "update"]


class RemoveAnalyticsField(Mutation):
    class Meta:
        serializer_class = serializers.RemoveAnalyticsFieldSerializer
        lookup_input_kwarg = "id"
        model_operations = ["update"]


class ReorderAnalyticsFields(Mutation):
    class Meta:
        serializer_class = serializers.ReorderAnalyticsFieldSerializer
        lookup_input_kwarg = "table"
        model_operations = ["update"]


class Mutation:
    save_analytics_table = SaveAnalyticsTable().Field()
    remove_analytics_table = RemoveAnalyticsTable().Field()

    save_analytics_field = SaveAnalyticsField().Field()
    remove_analytics_field = RemoveAnalyticsField().Field()
    reorder_analytics_fields = ReorderAnalyticsFields().Field()


class Query:
    all_analytics_tables = DjangoFilterConnectionField(
        AnalyticsTable,
        filterset_class=CollectionFilterSetFactory(
            filterset_class=filters.AnalyticsTableFilterSet,
            orderset_class=filters.AnalyticsTableOrderSet,
        ),
    )
    analytics_table = graphene.Field(AnalyticsTable, slug=String(required=True))

    all_analytics_fields = DjangoFilterConnectionField(
        AnalyticsField,
        filterset_class=CollectionFilterSetFactory(
            filterset_class=filters.AnalyticsFieldFilterSet,
            orderset_class=filters.AnalyticsFieldOrderSet,
        ),
    )

    @staticmethod
    def resolve_analytics_table(_, info, slug):
        return models.AnalyticsTable.objects.get(pk=slug)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.18 on 2019-02-08 10:30
from __future__ import unicode_literals

from django.db import migrations


def forwards_func(apps, schema_editor):
    Question = apps.get_model("caluma_form", "Question")
    Question.objects.filter(type="checkbox").update(type="multiple_choice")
    Question.objects.filter(type="radio").update(type="choice")


def reverse_func(apps, schema_editor):
    Question = apps.get_model("caluma_form", "Question")
    Question.objects.filter(type="multiple_choice").update(type="checkbox")
    Question.objects.filter(type="choice").update(type="radio")


class Migration(migrations.Migration):

    dependencies = [("caluma_form", "0005_auto_20190208_1016")]

    operations = [migrations.RunPython(forwards_func, reverse_func)]



import uuid
from functools import wraps

from django.contrib.postgres.fields import ArrayField
from django.contrib.postgres.indexes import GinIndex
from django.db import models, transaction
from django.utils.functional import cached_property
from localized_fields.fields import LocalizedField, LocalizedTextField
from minio import S3Error
from simple_history.models import HistoricalRecords

from ..caluma_core import models as core_models
from .storage_clients import client


class Form(core_models.SlugModel):
    name = LocalizedField(blank=False, null=False, required=False)
    description = LocalizedField(blank=True, null=True, required=False)
    meta = models.JSONField(default=dict)
    is_published = models.BooleanField(default=False)
    is_archived = models.BooleanField(default=False)
    questions = models.ManyToManyField(
        "Question", through="FormQuestion", related_name="forms"
    )
    source = models.ForeignKey(
        "self",
        blank=True,
        null=True,
        help_text="Reference this form has been copied from",
        related_name="+",
        on_delete=models.SET_NULL,
    )

    @classmethod
    def _get_all_raw_forms(cls, forms: list):
        return cls.objects.raw(
            """
            WITH RECURSIVE
                forms(slug) AS (
                    SELECT slug
                    FROM caluma_form_form
                    WHERE slug = ANY (%s)
                    UNION ALL
                    SELECT
                        all_subforms.subform_id AS slug
                    FROM forms JOIN all_subforms ON (all_subforms.form_id = forms.slug )
                ),
                all_subforms(form_id, subform_id) AS (
                    SELECT
                        caluma_form_formquestion.form_id,
                        COALESCE(caluma_form_question.row_form_id, caluma_form_question.sub_form_id) AS subform_id
                    FROM caluma_form_formquestion
                    JOIN caluma_form_question ON (caluma_form_formquestion.question_id = caluma_form_question.slug)
                    WHERE caluma_form_question.type IN ('form', 'table')
                )
            SELECT * FROM forms;
        """,
            [forms],
        )

    @classmethod
    def get_all_forms(cls, forms: list):  # pragma: no cover
        return Form.objects.filter(pk__in=cls._get_all_raw_forms(forms))

    @classmethod
    def get_all_questions(cls, forms: list):
        return Question.objects.filter(forms__in=cls._get_all_raw_forms(forms))

    class Meta:
        indexes = [GinIndex(fields=["meta"])]


class FormQuestion(core_models.NaturalKeyModel):
    form = models.ForeignKey("Form", on_delete=models.CASCADE)
    question = models.ForeignKey("Question", on_delete=models.CASCADE)
    sort = models.PositiveIntegerField(editable=False, db_index=True, default=0)

    def natural_key(self):
        return f"{self.form_id}.{self.question_id}"

    class Meta:
        ordering = ["-sort"]
        unique_together = ("form", "question")


class Question(core_models.SlugModel):
    TYPE_MULTIPLE_CHOICE = "multiple_choice"
    TYPE_INTEGER = "integer"
    TYPE_FLOAT = "float"
    TYPE_DATE = "date"
    TYPE_CHOICE = "choice"
    TYPE_TEXTAREA = "textarea"
    TYPE_TEXT = "text"
    TYPE_TABLE = "table"
    TYPE_FORM = "form"
    TYPE_FILES = "files"
    TYPE_DYNAMIC_CHOICE = "dynamic_choice"
    TYPE_DYNAMIC_MULTIPLE_CHOICE = "dynamic_multiple_choice"
    TYPE_STATIC = "static"
    TYPE_CALCULATED_FLOAT = "calculated_float"
    TYPE_ACTION_BUTTON = "action_button"

    TYPE_CHOICES = (
        TYPE_MULTIPLE_CHOICE,
        TYPE_INTEGER,
        TYPE_FLOAT,
        TYPE_DATE,
        TYPE_CHOICE,
        TYPE_TEXTAREA,
        TYPE_TEXT,
        TYPE_TABLE,
        TYPE_FORM,
        TYPE_FILES,
        TYPE_DYNAMIC_CHOICE,
        TYPE_DYNAMIC_MULTIPLE_CHOICE,
        TYPE_STATIC,
        TYPE_CALCULATED_FLOAT,
        TYPE_ACTION_BUTTON,
    )
    TYPE_CHOICES_TUPLE = [(type_choice, type_choice) for type_choice in TYPE_CHOICES]

    ACTION_COMPLETE = "complete"
    ACTION_SKIP = "skip"
    ACTION_CHOICES = (
        ACTION_COMPLETE,
        ACTION_SKIP,
    )

    COLOR_PRIMARY = "primary"
    COLOR_SECONDARY = "secondary"
    COLOR_DEFAULT = "default"
    COLOR_CHOICES = (
        COLOR_PRIMARY,
        COLOR_SECONDARY,
        COLOR_DEFAULT,
    )

    label = LocalizedField(blank=False, null=False, required=False)
    type = models.CharField(choices=TYPE_CHOICES_TUPLE, max_length=23)
    is_required = models.TextField(default="false")
    is_hidden = models.TextField(default="false")
    is_archived = models.BooleanField(default=False)
    placeholder = LocalizedField(blank=True, null=True, required=False)
    info_text = LocalizedField(blank=True, null=True, required=False)
    hint_text = LocalizedField(blank=True, null=True, required=False)
    static_content = LocalizedTextField(blank=True, null=True, required=False)
    configuration = models.JSONField(default=dict)
    meta = models.JSONField(default=dict)
    data_source = models.CharField(max_length=255, blank=True, null=True)
    options = models.ManyToManyField(
        "Option", through="QuestionOption", related_name="questions"
    )
    row_form = models.ForeignKey(
        Form,
        blank=True,
        null=True,
        related_name="+",
        help_text="Form that represents rows of a TableQuestion",
        on_delete=models.PROTECT,
    )
    sub_form = models.ForeignKey(
        Form,
        blank=True,
        null=True,
        related_name="+",
        help_text="Form referenced in a FormQuestion",
        on_delete=models.PROTECT,
    )

    source = models.ForeignKey(
        "self",
        blank=True,
        null=True,
        related_name="+",
        help_text="Reference this question has been copied from",
        on_delete=models.SET_NULL,
    )
    format_validators = ArrayField(
        models.CharField(max_length=255), blank=True, default=list
    )

    default_answer = models.OneToOneField(
        "Answer", on_delete=models.SET_NULL, related_name="+", null=True, blank=True
    )

    calc_expression = models.TextField(null=True, blank=True)
    calc_dependents = ArrayField(
        models.CharField(max_length=255, blank=True), default=list
    )

    @property
    def min_length(self):
        return self.configuration.get("min_length")

    @min_length.setter
    def min_length(self, value):
        self.configuration["min_length"] = value

    @property
    def max_length(self):
        return self.configuration.get("max_length")

    @max_length.setter
    def max_length(self, value):
        self.configuration["max_length"] = value

    @property
    def max_value(self):
        return self.configuration.get("max_value")

    @max_value.setter
    def max_value(self, value):
        self.configuration["max_value"] = value

    @property
    def min_value(self):
        return self.configuration.get("min_value")

    @min_value.setter
    def min_value(self, value):
        self.configuration["min_value"] = value

    @property
    def action(self):
        return self.configuration.get("action")

    @action.setter
    def action(self, value):
        self.configuration["action"] = value

    @property
    def color(self):
        return self.configuration.get("color")

    @color.setter
    def color(self, value):
        self.configuration["color"] = value

    @property
    def validate_on_enter(self):
        return self.configuration.get("validate_on_enter")

    @validate_on_enter.setter
    def validate_on_enter(self, value):
        self.configuration["validate_on_enter"] = value

    def empty_value(self):
        """Return empty value for this question type."""

        empties = {
            Question.TYPE_MULTIPLE_CHOICE: [],
            Question.TYPE_TABLE: [],
            Question.TYPE_DYNAMIC_MULTIPLE_CHOICE: [],
        }
        return empties.get(self.type, None)

    def __repr__(self):
        base = super().__repr__()
        return base[:-1] + f", type={self.type})"

    class Meta:
        indexes = [GinIndex(fields=["meta"])]


class QuestionOption(core_models.NaturalKeyModel):
    question = models.ForeignKey("Question", on_delete=models.CASCADE)
    option = models.ForeignKey("Option", on_delete=models.CASCADE)
    sort = models.PositiveIntegerField(editable=False, db_index=True, default=0)

    def natural_key(self):
        return f"{self.question_id}.{self.option_id}"

    class Meta:
        ordering = ["-sort"]
        unique_together = ("option", "question")


class Option(core_models.SlugModel):
    label = LocalizedField(blank=False, null=False, required=False)
    is_archived = models.BooleanField(default=False)
    meta = models.JSONField(default=dict)
    source = models.ForeignKey(
        "self",
        blank=True,
        null=True,
        related_name="+",
        help_text="Reference this option has been copied from",
        on_delete=models.SET_NULL,
    )

    class Meta:
        indexes = [GinIndex(fields=["meta"])]


class DocumentManager(models.Manager):
    @transaction.atomic
    def create_document_for_task(self, task, user):
        """Create a document for a given task."""
        if task.form_id is not None:
            # import domain logic here, in order to avoid recursive import error
            from .domain_logic import SaveDocumentLogic

            return SaveDocumentLogic.create({"form": task.form}, user=user)


class Document(core_models.UUIDModel):
    objects = DocumentManager()

    family = models.ForeignKey(
        "self",
        help_text="Family id which document belongs too.",
        null=True,
        on_delete=models.CASCADE,
        related_name="+",
    )
    form = models.ForeignKey(
        "caluma_form.Form", on_delete=models.DO_NOTHING, related_name="documents"
    )
    source = models.ForeignKey(
        "self",
        blank=True,
        null=True,
        help_text="Reference this document has been copied from",
        related_name="copies",
        on_delete=models.SET_NULL,
    )
    meta = models.JSONField(default=dict)

    def set_family(self, root_doc):
        """Set the family to the given root_doc.

        Apply the same to all row documents within this document,
        recursively.
        """
        if self.family != root_doc:
            self.family = root_doc
            self.save()

        table_answers = Answer.objects.filter(
            document=self,
            # TODO this is not REALLY required, as only table answers
            # have child documents anyhow. Leaving it out would avoid
            # a JOIN to the question table
            question__type=Question.TYPE_TABLE,
        )

        child_documents = AnswerDocument.objects.filter(
            answer__in=table_answers
        ).select_related("document")

        for answer_document in child_documents:
            answer_document.document.set_family(root_doc)

    def copy(self, family=None, user=None):
        """Create a copy including all its answers."""

        new_document = type(self).objects.create(
            form=self.form,
            meta=dict(self.meta),
            source=self,
            family=family,
            created_by_user=user.username if user else None,
            created_by_group=user.group if user else None,
            modified_by_user=user.username if user else None,
            modified_by_group=user.group if user else None,
        )
        if not family:
            family = new_document

        # copy answers
        for source_answer in self.answers.all():
            source_answer.copy(
                document_family=family, to_document=new_document, user=user
            )

        return new_document

    @cached_property
    def last_modified_answer(self):
        """Get most recently modified answer of the document.

        For root documents, we want to look through the whole document, while
        for row documents only the table-local answer is wanted.
        """
        if self.family != self:
            answers = self.answers
        else:
            answers = Answer.objects.filter(document__family=self)

        return answers.order_by("-modified_at").first()

    @property
    def modified_content_at(self):
        return getattr(self.last_modified_answer, "modified_at", None)

    @property
    def modified_content_by_user(self):
        return getattr(self.last_modified_answer, "modified_by_user", None)

    @property
    def modified_content_by_group(self):
        return getattr(self.last_modified_answer, "modified_by_group", None)

    def __repr__(self):
        return f"Document(form={self.form!r})"

    class Meta:
        indexes = [GinIndex(fields=["meta"])]


class QuestionTypeHistoricalModel(models.Model):
    """Extra fields for HistoricalAnswer."""

    history_question_type = models.CharField(
        choices=Question.TYPE_CHOICES_TUPLE, max_length=23
    )

    class Meta:
        abstract = True


class Answer(core_models.BaseModel):
    """Records an answer to a question of arbitrary type."""

    # We need to replicate the UUIDModel in order to register it as historical model.
    # Otherwise simple_history complains that the model is already registered,
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    question = models.ForeignKey(
        "caluma_form.Question", on_delete=models.DO_NOTHING, related_name="answers"
    )
    value = models.JSONField(null=True, blank=True)
    meta = models.JSONField(default=dict)
    document = models.ForeignKey(
        Document,
        on_delete=models.CASCADE,
        related_name="answers",
        null=True,
        blank=True,
    )
    documents = models.ManyToManyField(
        Document, through="AnswerDocument", related_name="+"
    )
    date = models.DateField(null=True, blank=True)

    # override history to add extra fields on historical model
    history = HistoricalRecords(
        inherit=True,
        history_user_id_field=models.CharField(null=True, max_length=150),
        history_user_setter=core_models._history_user_setter,
        history_user_getter=core_models._history_user_getter,
        bases=[QuestionTypeHistoricalModel],
    )

    def delete(self, *args, **kwargs):
        # Files need to be deleted in sequence (not via on_delete)
        # so the deletion code on the storage is properly triggered
        for file in self.files.all():
            file.delete()
        super().delete(args, kwargs)

    def create_answer_documents(self, documents):
        family = getattr(self.document, "family", None)
        document_ids = [document.pk for document in documents]

        for sort, document_id in enumerate(reversed(document_ids), start=1):
            ans_doc, created = AnswerDocument.objects.get_or_create(
                answer=self, document_id=document_id, defaults={"sort": sort}
            )
            if not created and ans_doc.sort != sort:
                ans_doc.sort = sort
                ans_doc.save()
            if created:
                # Already-existing documents are already in the family,
                # so we're updating only the newly attached rows
                ans_doc.document.set_family(family)

    def unlink_unused_rows(self, docs_to_keep):
        existing = AnswerDocument.objects.filter(answer=self).exclude(
            document__in=docs_to_keep
        )
        for ans_doc in list(existing.select_related("document")):
            # Set document to be its own family
            # TODO: Can/should we delete the detached documents?
            ans_doc.document.set_family(ans_doc.document)
            ans_doc.delete()

    def copy(self, document_family=None, to_document=None, user=None):
        new_answer, _ = type(self).objects.update_or_create(
            question=self.question,
            document=to_document,
            defaults={
                "value": self.value,
                "meta": dict(self.meta),
                "date": self.date,
                "created_by_user": user.username if user else None,
                "created_by_group": user.group if user else None,
                "modified_by_user": user.username if user else None,
                "modified_by_group": user.group if user else None,
            },
        )

        if self.question.type == Question.TYPE_FILES:
            for file in self.files.all():
                file.copy(to_answer=new_answer)

        if self.question.type in [
            Question.TYPE_DYNAMIC_CHOICE,
            Question.TYPE_DYNAMIC_MULTIPLE_CHOICE,
        ]:
            for dynamic_option in DynamicOption.objects.filter(
                document=self.document, question=self.question
            ):
                DynamicOption.objects.update_or_create(
                    document=to_document,
                    question=dynamic_option.question,
                    slug=dynamic_option.slug,
                    defaults={
                        "label": dynamic_option.label,
                        "created_by_user": user.username if user else None,
                        "created_by_group": user.group if user else None,
                    },
                )

        # TableAnswer: copy AnswerDocument too
        for answer_doc in AnswerDocument.objects.filter(answer=self):
            new_doc = answer_doc.document.copy(family=document_family, user=user)

            AnswerDocument.objects.create(
                answer=new_answer,
                document=new_doc,
                sort=answer_doc.sort,
                created_by_user=user.username if user else None,
                created_by_group=user.group if user else None,
                modified_by_user=user.username if user else None,
                modified_by_group=user.group if user else None,
            )
        return new_answer

    @property
    def selected_options(self):
        map = {
            Question.TYPE_CHOICE: (Option, {"slug": self.value}),
            Question.TYPE_MULTIPLE_CHOICE: (Option, {"slug__in": self.value}),
            Question.TYPE_DYNAMIC_CHOICE: (
                DynamicOption,
                {
                    "slug": self.value,
                    "question": self.question,
                    "document": self.document,
                },
            ),
            Question.TYPE_DYNAMIC_MULTIPLE_CHOICE: (
                DynamicOption,
                {
                    "slug__in": self.value,
                    "question": self.question,
                    "document": self.document,
                },
            ),
        }

        if self.question.type not in map:
            return None

        if not self.value:
            return []

        model, filters = map[self.question.type]
        queryset = model.objects.filter(**filters)

        def index_from_value(obj):
            return self.value.index(obj.slug)

        # Keep the same order as the value
        return sorted(queryset, key=index_from_value)

    def __repr__(self):
        return f"Answer(document={self.document!r}, question={self.question!r}, value={self.value!r})"

    class Meta:
        # a question may only be answerd once per document
        unique_together = ("document", "question")
        indexes = [models.Index(fields=["date"]), GinIndex(fields=["meta", "value"])]


def _ignore_missing_file(fn):
    """Ignore errors due to missing file.

    If no file is uploaded to the storage service, we want to ignore the resulting
    exceptions in order to still be able to modify/delete the File.
    """

    @wraps(fn)
    def wrapper(self, *args, **kwargs):
        try:
            return fn(self, *args, **kwargs)
        except S3Error as exc:
            if exc.code == "NoSuchKey":
                return
            raise

    return wrapper


class File(core_models.UUIDModel):
    name = models.CharField(max_length=255)

    answer = models.ForeignKey(
        Answer, on_delete=models.CASCADE, related_name="files", null=True, default=None
    )

    @_ignore_missing_file
    def _move_blob(self):
        # move the file on update
        # this makes sure it stays available when querying the history
        old_file = self.history.first()
        if not old_file:  # pragma: no cover
            # No history - cannot move object to preserve history.
            return
        new_name = f"{old_file.pk}_{old_file.name}"
        client.move_object(self.object_name, new_name)

    @_ignore_missing_file
    def _copy(self, new_object_name):
        client.copy_object(self.object_name, new_object_name)

    def copy(self, to_answer):
        copy = File.objects.create(name=self.name, answer=to_answer or self.answer)
        self._copy(copy.object_name)
        return copy

    def rename(self, new_name):
        from_object_name = self.object_name
        self.name = new_name
        client.move_object(from_object_name, self.object_name)

    def delete(self, *args, **kwargs):
        self._move_blob()
        super().delete(*args, **kwargs)

    def save(self, *args, **kwargs):
        if self.created_at:
            self._move_blob()
        super().save(*args, **kwargs)

    @property
    def object_name(self):
        return f"{self.pk}_{self.name}"

    @property
    def upload_url(self):
        return client.upload_url(self.object_name)

    @property
    def download_url(self):
        return client.download_url(self.object_name)

    @property
    def metadata(self):
        stat = client.stat_object(self.object_name)
        if not stat:
            # This could happen if no file was uploaded, but the client
            # still requests the property
            return None
        return {
            # There are some more metadata values, but we're
            # not showing everything here. Most of the remaining
            # stuff is not set, or not useful for clients anyway
            "bucket_name": stat.bucket_name,
            "content_type": stat.content_type,
            "etag": stat.etag,
            "last_modified": stat.last_modified.isoformat(),
            "metadata": dict(stat.metadata),
            "size": stat.size,
        }


class AnswerDocument(core_models.UUIDModel):
    answer = models.ForeignKey("Answer", on_delete=models.CASCADE)
    document = models.ForeignKey("Document", on_delete=models.CASCADE)
    sort = models.PositiveIntegerField(editable=False, db_index=True, default=0)

    class Meta:
        ordering = ["-sort"]
        unique_together = ("answer", "document")


class DynamicOption(core_models.UUIDModel):
    slug = models.CharField(max_length=255)
    label = LocalizedField(blank=False, null=False, required=False)
    document = models.ForeignKey("Document", on_delete=models.CASCADE)
    question = models.ForeignKey("Question", on_delete=models.CASCADE)

    class Meta:
        unique_together = ("slug", "document", "question")



# -*- coding: utf-8 -*-
# Generated by Django 1.11.18 on 2019-02-13 09:39
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("caluma_workflow", "0008_auto_20190208_1302")]

    operations = [
        migrations.AddField(
            model_name="case",
            name="closed_at",
            field=models.DateTimeField(
                blank=True,
                help_text="Time when case has either been canceled or completed",
                null=True,
            ),
        ),
        migrations.AddField(
            model_name="case",
            name="closed_by_group",
            field=models.CharField(blank=True, max_length=150, null=True),
        ),
        migrations.AddField(
            model_name="case",
            name="closed_by_user",
            field=models.CharField(blank=True, max_length=150, null=True),
        ),
        migrations.AddField(
            model_name="workitem",
            name="closed_at",
            field=models.DateTimeField(
                blank=True,
                help_text="Time when work item has either been canceled or completed",
                null=True,
            ),
        ),
        migrations.AddField(
            model_name="workitem",
            name="closed_by_group",
            field=models.CharField(blank=True, max_length=150, null=True),
        ),
        migrations.AddField(
            model_name="workitem",
            name="closed_by_user",
            field=models.CharField(blank=True, max_length=150, null=True),
        ),
    ]



import sys

from django.dispatch import Signal

module = sys.modules[__name__]

MODEL_ACTIONS = {
    "work_item": ["create", "complete", "cancel", "skip", "suspend", "resume", "redo"],
    "case": ["create", "complete", "cancel", "suspend", "resume", "reopen"],
}

# create "pre|post"_"create|complete|..."_"work_item|case" events in all given combinations
for obj, actions in MODEL_ACTIONS.items():
    for action in actions:
        for prefix in ["pre", "post"]:

            setattr(module, f"{prefix}_{action}_{obj}", Signal())



from functools import wraps

name_attr = "_dynamic_task_name"


def register_dynamic_task(dynamic_task_name):
    """Decorate function to register a dynamic task."""

    def wrapper(fn, *arg, **kwarg):
        setattr(fn, name_attr, dynamic_task_name)

        @wraps(fn)
        def wrapped(self, *args, **kwargs):
            return fn(self, *args, **kwargs)

        return wrapped

    return wrapper


class BaseDynamicTasks:
    """Basic dynamic tasks class to be extended by any dynamic tasks class implementation.

    In combination with the decorator `@register_dynamic_task` a dynamic task
    class can extend the JEXL `task` and `tasks` transform used in the `next`
    property of a task.

    A dynamic tasks class could look like this:

    >>> from caluma.caluma_workflow.dynamic_tasks import (
    ...     BaseDynamicTasks,
    ...     register_dynamic_task,
    ... )
    ...
    ...
    ... class CustomDynamicTasks(BaseDynamicTasks):
    ...     @register_dynamic_task("audit-forms")
    ...     def resolve_audit_forms(self, case, user, prev_work_item, context):
    ...         needs_extra_audit = case.document.answers.filter(
    ...             question_id="is-extra-complicated",
    ...             value="is-extra-complicated-yes"
    ...         ).exists()
    ...         next_tasks = ["audit"]
    ...
    ...         if needs_extra_audit:
    ...             next_tasks.append("extra-audit")
    ...
    ...         return next_tasks
    """

    def get_registered_dynamic_tasks(self):
        return filter(
            None,
            [
                getattr(getattr(self, methodname), name_attr, None)
                for methodname in dir(self)
            ],
        )

    def resolve(self, dynamic_task_name):
        for methodname in dir(self):
            method = getattr(self, methodname)

            if getattr(method, name_attr, None) == dynamic_task_name:
                return method



from django.db.models import Q

from ..caluma_core.visibilities import BaseVisibility, filter_queryset_for
from ..caluma_form import models as form_models
from ..caluma_form.schema import Answer, Document
from . import models
from .schema import Case, WorkItem


class AddressedGroups(BaseVisibility):
    """Only show case, work item and document to addressed users through group."""

    def get_visible_work_items(self, node, queryset, info):
        groups = info.context.user.groups
        return models.WorkItem.objects.filter(addressed_groups__overlap=groups)

    def get_visible_cases(self, node, queryset, info):
        work_items = self.get_visible_work_items(node, queryset, info)
        cases = models.Case.objects.filter(
            Q(work_items__in=work_items) | Q(parent_work_item__in=work_items)
        )
        return cases

    def get_visible_documents(self, node, queryset, info):
        work_items = self.get_visible_work_items(node, queryset, info)
        cases = self.get_visible_cases(node, queryset, info)
        families = form_models.Document.objects.filter(
            Q(work_item__in=work_items) | Q(case__in=cases)
        ).values("family")
        return form_models.Document.objects.filter(family__in=families)

    @filter_queryset_for(WorkItem)
    def filter_querset_for_work_item(self, node, queryset, info):
        work_items = self.get_visible_work_items(node, queryset, info)
        return queryset.filter(pk__in=work_items)

    @filter_queryset_for(Case)
    def filter_queryset_for_case(self, node, queryset, info):
        cases = self.get_visible_cases(node, queryset, info)
        return queryset.filter(pk__in=cases)

    @filter_queryset_for(Document)
    def filter_queryset_for_document(self, node, queryset, info):
        documents = self.get_visible_documents(node, queryset, info)
        return queryset.filter(pk__in=documents)

    @filter_queryset_for(Answer)
    def filter_queryset_for_answer(self, node, queryset, info):
        documents = self.get_visible_documents(node, queryset, info)
        return queryset.filter(document__in=documents)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.21 on 2019-06-21 05:40
from __future__ import unicode_literals

import logging

from django.db import migrations

logger = logging.getLogger(__name__)


def attach_form_sub_answers_to_main_document(apps, schema_editor):
    """Answers to sub-questions of FormQuestions share document with parent."""
    Answer = apps.get_model("caluma_form", "Answer")
    to_delete = []

    count = Answer.objects.filter(question__type="form").count()

    for i, a in enumerate(Answer.objects.filter(question__type="form")):
        logger.info(f"    migrating... {i+1}/{count}")

        if a.value_document:
            for answer in a.value_document.answers.all():
                logger.debug(
                    f"moving answer {answer.pk} from {answer.document.pk} to root {a.document.family}"
                )
                answer.document_id = a.document.family
                answer.save()

            logger.debug(
                f"mark unused value document for deletion: {a.value_document.pk}"
            )
            to_delete.append(a.value_document)
        logger.debug(f"mark unused form answer for deletion: {a.pk}")
        to_delete.append(a)

    for obj in to_delete:
        logger.debug(f"deleting {obj} {obj.pk}")
        obj.delete()


def undo_attach_form_sub_answers_to_main_document(apps, schema_editor):
    """Provide a dummy for reverting migrations."""
    logger.warning(
        "Reverting '0018_remove_answer_value_document_data_migration'"
        "doesn't nest the answer structure."
    )


class Migration(migrations.Migration):

    dependencies = [("caluma_form", "0017_auto_20190619_1320")]

    operations = [
        migrations.RunPython(
            attach_form_sub_answers_to_main_document,
            undo_attach_form_sub_answers_to_main_document,
        )
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.20 on 2019-04-16 18:35
from __future__ import unicode_literals

import localized_fields.fields.field
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("caluma_form", "0011_auto_20190411_0607")]

    operations = [
        migrations.AddField(
            model_name="question",
            name="info_text",
            field=localized_fields.fields.field.LocalizedField(
                blank=True, null=True, required=[]
            ),
        ),
        migrations.AddField(
            model_name="question",
            name="placeholder",
            field=localized_fields.fields.field.LocalizedField(
                blank=True, null=True, required=[]
            ),
        ),
    ]



from django.conf import settings
from django.urls import re_path

from caluma.caluma_core import views
from caluma.caluma_user.views import AuthenticationGraphQLView

urlpatterns = [
    re_path(
        "graphql/?",
        AuthenticationGraphQLView.as_view(graphiql=settings.DEBUG),
        name="graphql",
    ),
    re_path("healthz/?", views.health_check_status, name="healthz"),
]



import uuid
from io import StringIO

import requests
from django.conf import settings
from django.core import management
from watchman.decorators import check

from caluma.caluma_form import storage_clients
from caluma.caluma_logging.models import AccessLog


@check
def _check_pending_migrations(db_name):
    """Check database for pending migrations.

    Returns JSON mapping if no pending migrations, otherwise raises Exception.
    @check django-watchman decorator catches and handles exceptions.
    """
    # check for unapplied migrations
    out = StringIO()
    management.call_command("showmigrations", "--plan", stdout=out)
    plan = out.getvalue()
    if "[ ]" in plan:
        raise Exception("Database has unapplied migrations (migrate).")

    return {db_name: {"ok": True}}


@check
def _check_media_storage_service():
    """Check media storage service connection and operations."""
    # generate object name
    storage_client = storage_clients.client
    prefix = "healthz_tmp-object_"
    object_name = prefix + str(uuid.uuid4())

    try:
        # request upload url and upload
        upload_url = storage_client.upload_url(object_name)

        filename = "healthz_tmp-file.txt"
        content = "Test\nTest\n"
        files = {"file": (filename, content)}
        upload_request = requests.put(upload_url, files=files)
        assert upload_request.status_code == 200

        # stat object
        stat = storage_client.stat_object(object_name)
        assert stat.object_name == object_name

        # request download url and download file
        download_url = storage_client.download_url(object_name)

        download_request = requests.get(download_url)
        assert download_request.status_code == 200
        assert filename in download_request.text and content in download_request.text

        # remove object
        storage_client.remove_object(object_name)
        assert not storage_client.stat_object(object_name)

        return {"ok": True}

    finally:
        storage_client.remove_object(object_name)


@check
def _check_models(database):
    """Check model instantion and operations on database."""
    log = None
    try:
        # Instantiate model
        log = AccessLog.objects.create(status_code=22)
        assert (
            log in AccessLog.objects.all()
        ), "Failed to save model instance to database."

        # Retrieve object
        retrieved_log = AccessLog.objects.get(pk=log.pk)
        assert log == retrieved_log, "Unexpected object retrieved from database."
        assert (
            retrieved_log.status_code == log.status_code
        ), "Unexpected object data retrieved from database."

        # Update object
        log.status_code = 42
        log.save()
        retrieved_log = AccessLog.objects.get(pk=log.pk)
        assert log == retrieved_log, "Unexpected object retrieved from database."
        assert (
            retrieved_log.status_code == log.status_code
        ), "Unexpected object data retrieved from database."

        # Remove object
        log.delete()
        log = None
        assert (
            log not in AccessLog.objects.all()
        ), "Failed to delete model instance from database."

        return {database: {"ok": True}}

    finally:
        # Cleanup
        if log:  # pragma: no cover
            log.delete()


def check_media_storage_service():
    """Check media storage service connection and operations."""
    # Check if media storage service is available
    if not storage_clients.client:
        return {"media storage service (not configured)": {"ok": True}}

    return {"media storage service": _check_media_storage_service()}


def check_migrations():
    """Check available databases for unapplied migrations."""
    databases = sorted(settings.DATABASES)
    checked_databases = [_check_pending_migrations(db_name) for db_name in databases]

    return {"database migrations": checked_databases}


def check_models():
    """Check model instantiation."""
    databases = sorted(settings.DATABASES)
    checked_databases = [_check_models(db_name) for db_name in databases]

    return {"database models": checked_databases}



# Generated by Django 3.2.13 on 2022-06-28 08:16

import django.db.models.deletion
from django.db import migrations, models


def make_reverse_link(apps, schema_editor):
    Answer = apps.get_model("caluma_form.Answer")
    File = apps.get_model("caluma_form.File")
    HistoricalAnswer = apps.get_model("caluma_form.HistoricalAnswer")
    HistoricalFile = apps.get_model("caluma_form.HistoricalFile")

    # All files on current answers need to get the pointer file -> answer
    for file in File.objects.iterator():
        ans = Answer.objects.filter(file=file).first()
        file.answer = ans
        file.save()

    # Set answer on all HistoricalFiles
    for hfile in HistoricalFile.objects.iterator():
        hans = HistoricalAnswer.objects.filter(file_id=hfile.id).first()
        if not hans:  # pragma: no cover
            # really shouldn't happen
            continue
        hfile.answer_id = hans.id
        hfile.save()


def _rename_type(from_type, to_type, model):
    file_questions = model.objects.filter(type=from_type)
    file_questions.update(type=to_type)


def rename_file_type(apps, schema_editor):
    _rename_type("file", "files", apps.get_model("caluma_form.Question"))
    _rename_type("file", "files", apps.get_model("caluma_form.HistoricalQuestion"))


def rename_file_type_reverse(apps, schema_editor):
    _rename_type("files", "file", apps.get_model("caluma_form.Question"))
    _rename_type("files", "file", apps.get_model("caluma_form.HistoricalQuestion"))


class Migration(migrations.Migration):

    dependencies = [
        ("caluma_form", "0045_simple_history"),
    ]

    operations = [
        migrations.AddField(
            model_name="file",
            name="answer",
            field=models.ForeignKey(
                default=None,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="files",
                to="caluma_form.answer",
            ),
        ),
        migrations.AddField(
            model_name="historicalfile",
            name="answer",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                default=None,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="caluma_form.answer",
            ),
        ),
        migrations.AlterField(
            model_name="answer",
            name="file",
            field=models.OneToOneField(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="+",
                to="caluma_form.file",
            ),
        ),
        migrations.RunPython(make_reverse_link, migrations.RunPython.noop),
        migrations.RemoveField(
            model_name="answer",
            name="file",
        ),
        migrations.RemoveField(
            model_name="historicalanswer",
            name="file",
        ),
        migrations.AlterField(
            model_name="historicalanswer",
            name="history_question_type",
            field=models.CharField(
                choices=[
                    ("multiple_choice", "multiple_choice"),
                    ("integer", "integer"),
                    ("float", "float"),
                    ("date", "date"),
                    ("choice", "choice"),
                    ("textarea", "textarea"),
                    ("text", "text"),
                    ("table", "table"),
                    ("form", "form"),
                    ("files", "files"),
                    ("dynamic_choice", "dynamic_choice"),
                    ("dynamic_multiple_choice", "dynamic_multiple_choice"),
                    ("static", "static"),
                    ("calculated_float", "calculated_float"),
                    ("action_button", "action_button"),
                ],
                max_length=23,
            ),
        ),
        migrations.AlterField(
            model_name="historicalquestion",
            name="type",
            field=models.CharField(
                choices=[
                    ("multiple_choice", "multiple_choice"),
                    ("integer", "integer"),
                    ("float", "float"),
                    ("date", "date"),
                    ("choice", "choice"),
                    ("textarea", "textarea"),
                    ("text", "text"),
                    ("table", "table"),
                    ("form", "form"),
                    ("files", "files"),
                    ("dynamic_choice", "dynamic_choice"),
                    ("dynamic_multiple_choice", "dynamic_multiple_choice"),
                    ("static", "static"),
                    ("calculated_float", "calculated_float"),
                    ("action_button", "action_button"),
                ],
                max_length=23,
            ),
        ),
        migrations.AlterField(
            model_name="question",
            name="type",
            field=models.CharField(
                choices=[
                    ("multiple_choice", "multiple_choice"),
                    ("integer", "integer"),
                    ("float", "float"),
                    ("date", "date"),
                    ("choice", "choice"),
                    ("textarea", "textarea"),
                    ("text", "text"),
                    ("table", "table"),
                    ("form", "form"),
                    ("files", "files"),
                    ("dynamic_choice", "dynamic_choice"),
                    ("dynamic_multiple_choice", "dynamic_multiple_choice"),
                    ("static", "static"),
                    ("calculated_float", "calculated_float"),
                    ("action_button", "action_button"),
                ],
                max_length=23,
            ),
        ),
        migrations.RunPython(rename_file_type, rename_file_type_reverse),
    ]



import base64
import functools
import hashlib

import requests
from django.conf import settings
from django.core.cache import cache
from django.core.exceptions import ImproperlyConfigured
from django.http.response import HttpResponse
from django.utils.encoding import force_bytes, smart_str
from django.utils.module_loading import import_string
from graphene_django.views import GraphQLView, HttpError
from rest_framework.authentication import get_authorization_header

from caluma.caluma_user import models


class HttpResponseUnauthorized(HttpResponse):
    status_code = 401


class AuthenticationGraphQLView(GraphQLView):
    def get_bearer_token(self, request):
        auth = get_authorization_header(request).split()
        header_prefix = "Bearer"

        if not auth:
            return None

        if smart_str(auth[0].lower()) != header_prefix.lower():
            raise HttpError(HttpResponseUnauthorized("No Bearer Authorization header"))

        if len(auth) == 1:
            msg = "Invalid Authorization header. No credentials provided"
            raise HttpError(HttpResponseUnauthorized(msg))
        elif len(auth) > 2:
            msg = (
                "Invalid Authorization header. Credentials string should "
                "not contain spaces."
            )
            raise HttpError(HttpResponseUnauthorized(msg))

        return auth[1]

    def get_userinfo(self, token):
        response = requests.get(
            settings.OIDC_USERINFO_ENDPOINT,
            verify=settings.OIDC_VERIFY_SSL,
            headers={"Authorization": f"Bearer {smart_str(token)}"},
        )
        response.raise_for_status()
        return response.json()

    def get_introspection(self, token):
        basic = base64.b64encode(
            f"{settings.OIDC_INTROSPECT_CLIENT_ID}:{settings.OIDC_INTROSPECT_CLIENT_SECRET}".encode(
                "utf-8"
            )
        ).decode()
        headers = {
            "Authorization": f"Basic {basic}",
            "Content-Type": "application/x-www-form-urlencoded",
        }
        response = requests.post(
            settings.OIDC_INTROSPECT_ENDPOINT,
            verify=settings.OIDC_VERIFY_SSL,
            headers=headers,
            data={"token": token},
        )
        response.raise_for_status()
        return response.json()

    def _oidc_user(self, *args, **kwargs):
        factory = import_string(settings.CALUMA_OIDC_USER_FACTORY)
        return factory(*args, **kwargs)

    def get_user(self, request):
        token = self.get_bearer_token(request)

        if token is None:
            return models.AnonymousUser()

        if not settings.OIDC_USERINFO_ENDPOINT:
            raise ImproperlyConfigured(
                'Token was provided, but "OIDC_USERINFO_ENDPOINT" is not set.'
            )

        userinfo_method = functools.partial(self.get_userinfo, token=token)
        # token might be too long for key so we use hash sum instead.
        hashsum_token = hashlib.sha256(force_bytes(token)).hexdigest()

        try:
            userinfo = cache.get_or_set(
                f"authentication.userinfo.{hashsum_token}",
                userinfo_method,
                timeout=settings.OIDC_BEARER_TOKEN_REVALIDATION_TIME,
            )
            return self._oidc_user(token=token, userinfo=userinfo)
        except requests.HTTPError as e:
            try:
                if (
                    e.response.status_code in [401, 403]
                    and settings.OIDC_INTROSPECT_ENDPOINT
                ):
                    introspect_method = functools.partial(
                        self.get_introspection, token=token
                    )
                    introspection = cache.get_or_set(
                        f"authentication.introspect.{hashsum_token}",
                        introspect_method,
                        timeout=settings.OIDC_BEARER_TOKEN_REVALIDATION_TIME,
                    )
                    if "client_id" not in introspection:
                        response = HttpResponse(status=401)
                        raise HttpError(response)
                    return self._oidc_user(token=token, introspection=introspection)
                else:
                    raise e
            except requests.HTTPError as internal_exception:
                # convert request error to django http error
                response = HttpResponse()
                response.status_code = internal_exception.response.status_code
                raise HttpError(response, message=str(internal_exception))

    def dispatch(self, request, *args, **kwargs):
        try:
            request.user = self.get_user(request)
            return super().dispatch(request, *args, **kwargs)
        except HttpError as e:
            response = e.response
            response["Content-Type"] = "application/json"
            response.content = self.json_encode(
                request, {"errors": [self.format_error(e)]}
            )
            return response



# -*- coding: utf-8 -*-
# Generated by Django 1.11.20 on 2019-05-06 06:55
from __future__ import unicode_literals

import django.contrib.postgres.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("caluma_form", "0014_auto_20190429_0910")]

    operations = [
        migrations.AddField(
            model_name="question",
            name="format_validators",
            field=django.contrib.postgres.fields.ArrayField(
                base_field=models.CharField(max_length=255),
                blank=True,
                default=list,
                size=None,
            ),
        )
    ]



import functools
import logging

from django.core.cache import cache

logger = logging.getLogger(__name__)


def data_source_cache(timeout=None):
    def decorator(method):
        @functools.wraps(method)
        def handle_cache(self, *args, **kwargs):
            key = f"data_source_{type(self).__name__}"

            get_data_method = functools.partial(method, self, *args, **kwargs)
            return cache.get_or_set(key, get_data_method, timeout)

        return handle_cache

    return decorator



# Generated by Django 3.2.12 on 2022-03-31 14:24

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("caluma_analytics", "0002_analytics_pivot_table"),
    ]

    operations = [
        migrations.AlterField(
            model_name="analyticstable",
            name="starting_object",
            field=models.CharField(
                choices=[
                    ("cases", "Cases"),
                    ("work_items", "Work Items"),
                    ("documents", "documents"),
                ],
                max_length=250,
            ),
        ),
        migrations.AlterField(
            model_name="historicalanalyticstable",
            name="starting_object",
            field=models.CharField(
                choices=[
                    ("cases", "Cases"),
                    ("work_items", "Work Items"),
                    ("documents", "documents"),
                ],
                max_length=250,
            ),
        ),
    ]



# Generated by Django 3.2.13 on 2022-04-27 09:04

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("caluma_workflow", "0029_task_continue_async"),
    ]

    operations = [
        migrations.AlterField(
            model_name="case",
            name="status",
            field=models.CharField(
                choices=[
                    ("running", "Case is running and work items need to be completed."),
                    ("completed", "Case is done."),
                    ("canceled", "Case is canceled."),
                    ("suspended", "Case is suspended."),
                ],
                db_index=True,
                max_length=50,
            ),
        ),
        migrations.AlterField(
            model_name="historicalcase",
            name="status",
            field=models.CharField(
                choices=[
                    ("running", "Case is running and work items need to be completed."),
                    ("completed", "Case is done."),
                    ("canceled", "Case is canceled."),
                    ("suspended", "Case is suspended."),
                ],
                db_index=True,
                max_length=50,
            ),
        ),
        migrations.AlterField(
            model_name="historicaltask",
            name="type",
            field=models.CharField(
                choices=[
                    ("simple", "Task which can simply be marked as completed."),
                    (
                        "complete_workflow_form",
                        "Task to complete a defined workflow form.",
                    ),
                    ("complete_task_form", "Task to complete a defined task form."),
                ],
                max_length=50,
            ),
        ),
        migrations.AlterField(
            model_name="historicalworkitem",
            name="status",
            field=models.CharField(
                choices=[
                    ("ready", "Work item is ready to be processed."),
                    ("completed", "Work item is done."),
                    ("canceled", "Work item is canceled."),
                    ("skipped", "Work item is skipped."),
                    ("suspended", "Work item is suspended."),
                    ("redo", "Work item has been marked for redo."),
                ],
                db_index=True,
                max_length=50,
            ),
        ),
        migrations.AlterField(
            model_name="task",
            name="type",
            field=models.CharField(
                choices=[
                    ("simple", "Task which can simply be marked as completed."),
                    (
                        "complete_workflow_form",
                        "Task to complete a defined workflow form.",
                    ),
                    ("complete_task_form", "Task to complete a defined task form."),
                ],
                max_length=50,
            ),
        ),
        migrations.AlterField(
            model_name="workitem",
            name="status",
            field=models.CharField(
                choices=[
                    ("ready", "Work item is ready to be processed."),
                    ("completed", "Work item is done."),
                    ("canceled", "Work item is canceled."),
                    ("skipped", "Work item is skipped."),
                    ("suspended", "Work item is suspended."),
                    ("redo", "Work item has been marked for redo."),
                ],
                db_index=True,
                max_length=50,
            ),
        ),
    ]



# Generated by Django 3.2.13 on 2022-06-09 11:26

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("caluma_workflow", "0030_remove_choices_char_field"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="historicalcase",
            options={
                "get_latest_by": ("history_date", "history_id"),
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical case",
                "verbose_name_plural": "historical cases",
            },
        ),
        migrations.AlterModelOptions(
            name="historicalflow",
            options={
                "get_latest_by": ("history_date", "history_id"),
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical flow",
                "verbose_name_plural": "historical flows",
            },
        ),
        migrations.AlterModelOptions(
            name="historicaltask",
            options={
                "get_latest_by": ("history_date", "history_id"),
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical task",
                "verbose_name_plural": "historical tasks",
            },
        ),
        migrations.AlterModelOptions(
            name="historicaltaskflow",
            options={
                "get_latest_by": ("history_date", "history_id"),
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical task flow",
                "verbose_name_plural": "historical task flows",
            },
        ),
        migrations.AlterModelOptions(
            name="historicalworkflow",
            options={
                "get_latest_by": ("history_date", "history_id"),
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical workflow",
                "verbose_name_plural": "historical workflows",
            },
        ),
        migrations.AlterModelOptions(
            name="historicalworkitem",
            options={
                "get_latest_by": ("history_date", "history_id"),
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical work item",
                "verbose_name_plural": "historical work items",
            },
        ),
        migrations.AlterField(
            model_name="historicalcase",
            name="history_date",
            field=models.DateTimeField(db_index=True),
        ),
        migrations.AlterField(
            model_name="historicalflow",
            name="history_date",
            field=models.DateTimeField(db_index=True),
        ),
        migrations.AlterField(
            model_name="historicaltask",
            name="history_date",
            field=models.DateTimeField(db_index=True),
        ),
        migrations.AlterField(
            model_name="historicaltaskflow",
            name="history_date",
            field=models.DateTimeField(db_index=True),
        ),
        migrations.AlterField(
            model_name="historicalworkflow",
            name="history_date",
            field=models.DateTimeField(db_index=True),
        ),
        migrations.AlterField(
            model_name="historicalworkitem",
            name="history_date",
            field=models.DateTimeField(db_index=True),
        ),
    ]
